#!/bin/bash
# Performs full analysis of model stored in VAM file

SRCDIR=$(dirname $0)
TOPDIR=`realpath "$SRCDIR/.."`
BINDIR=`realpath "$TOPDIR/bin"`
VER=`cat "$TOPDIR/VERSION"`

if [ "$1" = "" ]; then
    echo "HADES version ${VER}."
    echo "Usage: analyse-vam.sh name [ --pc var ] [ --no-simpl ] [ --no-dfa ]"
    echo "           [ --no-micro ] [ --no-con ] [ --no-raw ] [ --silent ]"
    echo
    echo "    name          Name of VAM file."
    echo "    --pc var      Specification of the name of PC register (default \"v8\")."
    echo "    --no-simpl    No simplification of the input file (not recommended)."
    echo "    --no-dfa      Turns off data-flow analysis, looks for"
    echo "                  file <name>_anot.vam instead."
    echo "    --no-micro    Do not compute possible states of microarch. registers."
    echo "    --no-con      Turns off consistency analysis."
    echo "    --no-raw      Bypass RAW hazard analysis."
    echo "    --silent      Silent mode. Only performs time meausurements."
    echo 
    exit 1
fi

# Input file
IN=$1
shift

PC="v8"
SIMPL="YES"
DFA="YES"
MIC="YES"
CON="YES"
RAW="YES"
SIL="/dev/stdout"

while [[ $# > 0 ]]; do
    key="$1"

    case $key in
        --pc)
            PC="$1"
            shift
        ;;
        --no-simpl)
            SIMPL=NO
        ;;
        --no-dfa)
            DFA=NO
        ;;
        --no-micro)
            MIC=NO
        ;;
        --no-con)
            CON=NO
        ;;
        --no-raw)
            RAW=NO
        ;;
        --silent)
            SIL="/dev/null"
        ;;
        *)
            # unknown option
            echo "Warning: Unknown option ${1}."
        ;;
    esac
    shift
done

export PATH=$BINDIR:$PATH

FILE=$(basename $IN)
NAME="${FILE%%.*}"

PY=python
RTLMANIP=rtlmanip
TIME="/usr/bin/time --format=%e"
AWK=gawk

if [ "$SIMPL" = "YES" ]; then
    echo "Invoking rtlmanip for simplification ..."
    $TIME -o "${NAME}-simpl.time" $RTLMANIP -q "(simplify) (unifyselectors) (save \"${NAME}_simple.vam\")" $IN > $SIL || exit 1
    #$TIME -o "${NAME}-simpl.time" $RTLMANIP -q "(unifyselectors) (simplify) (save \"${NAME}_simple.vam\")" $IN > $SIL || exit 1
fi

if [ "$DFA" = "YES" ]; then
    echo "Performing data-flow analysis ..."
    $TIME -o "${NAME}-dfa.time" $RTLMANIP -q "(dataflow (select register $PC)) (save \"${NAME}_anot.vam\")" "${NAME}_simple.vam" > $SIL || exit 1
fi

if [ "$MIC" = "YES" ]; then
    echo "Computing states of micro-architectural registers ..."
    $TIME -o "${NAME}-micro.time" $PY "$SRCDIR/micro.py" "${NAME}_anot.vam" "${NAME}_anot.assert" > $SIL || exit 1
fi

if [ "$CON" = "YES" ]; then
    echo "Checking consistency ..."
    $TIME -o "${NAME}-con.time" $PY "$SRCDIR/consistency.py" "${NAME}_anot.vam" "${NAME}_anot.assert" "${NAME}-con.tstat" > $SIL || exit 1
fi

if [ "$RAW" = "YES" ]; then
    echo "Detecting possible hazards ..."
    $TIME -o "${NAME}-raw.time" $PY "$SRCDIR/hazard.py" "${NAME}_anot.vam" "${NAME}_anot.assert" "${NAME}.stat" "${NAME}-raw.tstat" > $SIL || exit 1
fi

# Compute total time from acquired times 
rm -f "${NAME}-total.time"
cat ${NAME}-*.time | awk '{s+=$1} END {print s}' > "${NAME}-total.time"

echo "simpl;dfa;micro;con;hazard;total;raw_hcs;war_hcs;waw_hcs;ctl_hcs;tot_hcs;con_rtl;con_smt;con_hades;haz_rtl;haz_smt;haz_armc;haz_hades" > "${NAME}.csv"
for w in simpl dfa micro con raw total; do
    if [ -f "${NAME}-${w}.time" ]; then
        cat ${NAME}-${w}.time | tr "\n" ";" >> "${NAME}.csv"
    else
        echo -n "DN;" >> "${NAME}.csv"
    fi
done

# No. of hazard cases
HCS=`$AWK -F';' 'NR>1{ts[$1]+=$3;tot+=$3} END { print (ts["RAW"]?ts["RAW"]:"0") ";" (ts["WAR"]?ts["WAR"]:"0") ";" (ts["WAW"]?ts["WAW"]:"0") ";" (ts["CTL"]?ts["CTL"]:"0") ";" (tot?tot:"0") ";"}' "${NAME}.stat"`
echo -n $HCS >> ${NAME}.csv

# Detailed time statistics of consistency checking phase
CON_TIME=`$AWK -F';' 'NR>1{ts[$1]+=$2} END { print (ts["rtlmanip"]?ts["rtlmanip"]:"0") ";" (ts["smt"]?ts["smt"]:"0") ";"}' "${NAME}-con.tstat"`
echo -n $CON_TIME >> ${NAME}.csv

if [ -f "${NAME}-con.time" ]; then
    CON_TOT=`cat ${NAME}-con.time | tr -d "\n"`

    CON_HADES=`$AWK -F';' -v total="$CON_TOT" 'NR>1{tot+=$2} END { print (tot?total-tot:"0") ";"}' "${NAME}-con.tstat"`
else
    CON_HADES="DN;"
fi
echo -n $CON_HADES >> ${NAME}.csv

# Detailed time statistics of hazard look-up phase
HAZ_TIME=`$AWK -F';' 'NR>1{ts[$1]+=$2} END { print (ts["rtlmanip"]?ts["rtlmanip"]:"0") ";" (ts["smt"]?ts["smt"]:"0") ";" (ts["armc"]?ts["armc"]:"0") ";"}' "${NAME}-raw.tstat"`
echo -n $HAZ_TIME >> ${NAME}.csv

if [ -f "${NAME}-raw.time" ]; then
    HAZ_TOT=`cat ${NAME}-raw.time | tr -d "\n"`

    HAZ_HADES=`$AWK -F';' -v total="$HAZ_TOT" 'NR>1{tot+=$2} END { print (tot?total-tot:"0")}' "${NAME}-raw.tstat"`
else
    HAZ_HADES="DN"
fi
echo $HAZ_HADES >> ${NAME}.csv


