#!/usr/bin/python
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Name:         common.py
# Description:  Shared utils.
# Author:       Lukas Charvat 27.3.2014
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import sys
import os
import re
import time

from collections import deque
import subprocess

# SMT solver (e.g., z3, boolector, ...)
SMTTOOL  = "z3"
ARMCTOOL = "artmc-rtl.sh"

TREG    = "reg"
TSIG    = "sig"
TMEM    = "port"
TFNC    = "fnode"
TMEMORY = "mem"

TLREG   = "register"
TLSIG   = "signal"
TLMEM   = "memport"
TLFNC   = "fnode"
TLMEMORY= "memory"

TMEMRO  = "ro"
TMEMWO  = "wo"

SMTERR      = -1
SMTUNSAT    = 0
SMTSAT      = 1

ARMCERR     = -1
ARMCHOLD    = 1 
ARMCBROKEN  = 0 

CST     = "stall"
CEN     = "enable"
CCL     = "clear"
TD      = "data-in"
TQ      = "data-out"
TADDR   = "address"

COI_BCK = "coib"
COI_FWD = "coi"

RTLMANIP = "rtlmanip"

# rtlmanip subprocess
gRtlProc = None
gRtlFile = None
query_memo = {}

# time statistics
g_rtlmanip_time = 0
g_smt_time      = 0
g_armc_time     = 0

def buildStdQuery():
    cmd = ""
    cmd = addCmd(cmd, "(print 1)")
    cmd = addDelim(cmd)
    return cmd

def getLongType(aType):
    arr = { TREG : TLREG, TSIG : TLSIG, TMEM : TLMEM, TFNC : TLFNC }
    return arr[aType]

gNumDelim = "424242424242"

def finish():
    global gRtlProc
    global gRtlFile

    if gRtlProc != None:
        gRtlProc.communicate("(quit)\n")

    gRtlProc = None
    gRtlFile = None

def query(aFile, aCmd):
    global gRtlProc
    global gRtlFile
    global query_memo
    global g_rtlmanip_time

    is_expensive = False
    if "smtlib-tag-prime" in aCmd:
        is_expensive = True
        val_memo = query_memo.get((aFile, aCmd), None)
        if val_memo != None:
            return val_memo

    if gRtlFile != aFile:
        if gRtlProc != None:
            gRtlProc.communicate("(quit)\n")
        gRtlProc = None
        gRtlFile = None
 
    if gRtlProc == None:
        gRtlFile = aFile

        # Mark current time
        t_start = time.time()

        gRtlProc = subprocess.Popen("%s %s" % (RTLMANIP, aFile), shell=True, \
            stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        # Get consumed time
        g_rtlmanip_time += time.time() - t_start

        if gRtlProc == None:
            error("Cannot create 'rtlmanip' subprocess!")
            exit(1)

    #info("=== Command ==========================================")
    #info(aCmd)

    # Mark current time
    t_start = time.time()

    # Execute command
    gRtlProc.stdin.write(aCmd)
    gRtlProc.stdin.write("(print 565656)\n")

    # Get consumed time
    g_rtlmanip_time += time.time() - t_start

    ret = ""
    line = gRtlProc.stdout.readline()
    while line and line != "565656\n":
        #print "l '%s'" % line
        if len(line) < 4 or line[:4] != "rtl>":
            ret += line
        line = gRtlProc.stdout.readline()

    #print "...", ret
    n = ret.find(gNumDelim)
    if n == -1:
        error("Cannot obtain result from 'rtlmanip'!")
        exit(1)

    ret = ret[n + len(gNumDelim) + 1:]
    #info("=== Result ===========================================")
    #info(ret)

    if is_expensive:
        query_memo[(aFile, aCmd)] = ret
    return ret

def addDelim(aCmd):
    return aCmd + "(print " + gNumDelim + ")\n"

def addCmd(aLine, aCmd):
    return aLine + aCmd + "\n"

def isIdChar(aChar):
    return aChar.isalnum() or aChar == "_"

RDNAME  = 0
SKIP    = 1

def parseSet(aSet):
    if (aSet.strip()[:4] != "(set"):
        return []
    else:
        i = 4
        d = 0
        s = RDNAME
        arr = []
        while (i < len(aSet)):
            c = aSet[i]
            if s == RDNAME:
                if c == "(":
                    t = ""
                    i = i + 1
                    while aSet[i].isspace():
                        i = i + 1
                    while isIdChar(aSet[i]):
                        t = t + aSet[i]
                        i = i + 1
                    stype = t
                    t = ""
                    while aSet[i].isspace():
                        i = i + 1
                    while isIdChar(aSet[i]):
                        t = t + aSet[i]
                        i = i + 1
                    sname = t
                    arr.append((sname, stype))
                    s = SKIP 
            elif s == SKIP:
                if c == "(":
                    d = d + 1
                elif c == ")":
                    if d == 0:
                        s = RDNAME
                    else:
                        d = d - 1
            i = i + 1
        return arr

def querySet(aFile, cmd):
    return parseSet(query(aFile, cmd))

def parseInt(aScalar):
    try:
        return int(aScalar)
    except ValueError:
        return None

def queryInt(aFile, cmd):
    return parseInt(query(aFile, cmd))

def getPortDir(aFile, aName):
    cmd = ""
    cmd = addCmd(cmd, "(print 1)")
    cmd = addDelim(cmd)
    cmd = addCmd(cmd, "(select memport " + aName + ")")
    q = query(aFile, cmd)
    if q[:5] != "(port":
        return None

    i = q.find("dir")
    i = i + len("dir") + 1
    t = ""
    while (q[i].isalpha()):
        t = t + q[i]
        i = i + 1 
    return t   

def getControlSigs(aFile, aName, aType):
    if aType == TREG:
        sigs = getPipeControlSigs(aFile, aName)
    else:
        sigs = []

    cmd = ""
    cmd = addCmd(cmd, "(print 1)")
    cmd = addDelim(cmd)
    cmd = addCmd(cmd, "(select " + getLongType(aType) + " " + aName + ")")
    q = query(aFile, cmd)
    if q[:4] != "(reg" and q[:5] != "(port":
        return sigs

    i = q.find("we")
    i = i + len("we") + 1
    t = ""
    while (q[i].isalnum()):
        t = t + q[i]
        i = i + 1
    sigs.append(t)

    i = q.find("en")
    i = i + len("en") + 1
    t = ""
    while (q[i].isalnum()):
        t = t + q[i]
        i = i + 1
    sigs.append(t)

    return sigs  

def getPipeControlSigs(aFile, aName):
    cmd = ""
    cmd = addCmd(cmd, "(print 1)")
    cmd = addDelim(cmd)
    cmd = addCmd(cmd, "(select register " + aName + ")")
    q = query(aFile, cmd)
    if q[:4] != "(reg":
        return []

    sigs = []

    i = q.find("stall")
    if i != -1:
        i = i + len("stall") + 1
        t = ""
        while (q[i].isalnum()):
            t = t + q[i]
            i = i + 1
        sigs.append(t)

    i = q.find("clr")
    if i != -1:
        i = i + len("clr") + 1
        t = ""
        while (q[i].isalnum()):
            t = t + q[i]
            i = i + 1
        sigs.append(t)

    return sigs 

def getComment(aFile, aName, aType):
    cmd = ""
    cmd = addCmd(cmd, "(print 1)")
    cmd = addDelim(cmd)
    cmd = addCmd(cmd, "(select " + getLongType(aType) + " " + aName + ")")
    q = query(aFile, cmd)

    i = q.find("comment")
    if i == -1: 
        return ""

    i = i + len("comment") + 2
    t = ""
    while q[i] != '"':
        t = t + q[i]
        i = i + 1

    return t 

def isPipe(aFile, aName, aType):
    if aType == TREG:
        return len(getPipeControlSigs(aFile, aName)) > 0
    else:
        return False

def isArch(aFile, aName, aType):
    if aType == TMEM:
        return True
    elif aType == TREG:
        cmd = ""
        cmd = addCmd(cmd, "(print 1)")
        cmd = addDelim(cmd)
        cmd = addCmd(cmd, "(select " + getLongType(aType) + " " + aName + ")")
        q = query(aFile, cmd)

        i = q.find("architectural")
        if i == -1: 
            return False
        else:
            return True
    else:
        return False

def isMicroArch(aFile, aName, aType):
    if aType == TREG:
        return (not isPipe(aFile, aName, aType)) and (not isArch(aFile, aName, aType))
    else:
        return False

def getArchStorages(aFile):
    cmd = ""
    cmd = addCmd(cmd, "(print 1)")
    cmd = addDelim(cmd)
    cmd = addCmd(cmd, "(set i where (in i (union registers memports)))")
    q = query(aFile, cmd)
    qlist = parseSet(q)
   
    result = []
    for item in qlist:
        if isArch(aFile, item[0], item[1]):
            result.append(item)

    return result

def smtlibExport(aStr):
    ret = ";;; Begin of SMTLib export\n"
    ret = ret + aStr + "\n"
    ret = ret + ";;; End of SMTLib export\n"
    return ret

def parseSmtlibExport(aStr, aComment=""):
    lines = aStr.split("\n")
    inside = False
    result = ""
    for line in lines:
        if line == ";;; Begin of SMTLib export":
            inside = True
            continue
        if line == ";;; End of SMTLib export":
            inside = False
            continue

        if line.strip() != "" and inside:
            result = result + line + "\n"
    result = (";;; %s\n" % aComment) + result
    return result

def runsmt(aStr):
    global g_smt_time

    # Export
    h = open("last_run.smt2", "w")
    h.write(aStr)
    h.close()

    # Mark current time
    t_start = time.time()

    # Run command
    cmdline = "%s last_run.smt2" % SMTTOOL
    stream = os.popen(cmdline)
    result = stream.read()

    # Get consumed time
    g_smt_time += time.time() - t_start

    if result.find("error") != -1:
        print result
        return SMTERR
    elif result.find("unsat") != -1:
        return SMTUNSAT
    elif result.find("sat") != -1:
        return SMTSAT
    else:
        #print result
        return SMTERR

def smtsat(aStr):
    aStr += "\n; Pre-printed command for possible debugging\n"
    aStr += ";(apply ctx-solver-simplify) ; simplifies whole instance (Z3 only)\n"
    aStr += "(check-sat)\n(exit)\n"

    return runsmt(aStr)

def smtGetBvModel(aStr):
    global g_smt_time

    # Export
    h = open("last_run.smt2", "w")
    h.write(aStr)
    h.close()

    # Mark current time
    t_start = time.time()

    # Run command
    cmdline = "%s last_run.smt2" % SMTTOOL
    t1 = time.time()
    stream = os.popen(cmdline)
    result = stream.read()

    # Get consumed time
    g_smt_time += time.time() - t_start

    if result.find("unsat") != -1:
        return None
    elif result.find("error") != -1:
        error("SmtGetBV error")
        exit(1)
    elif result.find("sat") != -1:
        lines = result.splitlines()
        line = " ".join(lines)
        finds = re.findall("\(define-fun (\w+) [^#]*(#\w[0-9a-fA-F]+)\)", line)
        return finds
    else:
        error("SmtGetBV error")
        return None

def runarmc(aFile):
    global g_armc_time

    # Mark current time
    t_start = time.time()

    # Run command
    cmdline = "%s %s" % (ARMCTOOL, aFile)
    stream = os.popen(cmdline)
    result = stream.read()

    # Get consumed time
    g_armc_time += time.time() - t_start

    if result.find("error") != -1:
        print result
        return ARMCERR
    elif result.find("Property holds") != -1:
        return ARMCHOLD
    elif result.find("Property broken") != -1:
        return ARMCBROKEN
    else:
        #print result
        return ARMCERR

def smtrun2str(aRun):
    arr = { SMTERR : "error", SMTUNSAT : "unsat", SMTSAT : "sat" }
    return arr[aRun]

def armcrun2str(aRun):
    arr = { ARMCERR : "error", ARMCHOLD : "holds", ARMCBROKEN : "broken" }
    return arr[aRun]

def warning(aStr):
    print >> sys.stdout, "W:", aStr 

def info(aStr):
    print >> sys.stdout, "I:", aStr 

def error(aStr):
    print >> sys.stdout, "E:", aStr

def questionYesNo(aStr, aDefAns=False):
    print >> sys.stdout, "Q:", aStr   
    ans = raw_input()
    return ans.lower() == "y" or (ans == "" and aDefAns)

def assertion(aCond, aMsg):
    if not aCond:
        error("Assertion: %s failed" % aMsg)

def getCtrl(aFile, aObj, aCtrl):
    cmd = buildStdQuery()
    cmd = addCmd(cmd, "(%s (select %s %s))" % (aCtrl, getLongType(aObj[1]), aObj[0]))
    q = query(aFile, cmd)
    qlist = parseSet(q)
    if len(qlist) > 0:
        return qlist[0]
    else:
        return None

def simpleQuery(aFile, aQuery):
    cmd = buildStdQuery()
    cmd = addCmd(cmd, aQuery)
    q = query(aFile, cmd)

    return q

def obj(aSig, aType):
    return (aSig, aType)

def getName(aObj):
    return aObj[0]

def getType(aObj):
    return aObj[1]

def getParentMemory(aFile, aPort):
    if getType(aPort) != TMEM:
        error("Expected memport, got %s." % str(aPort))
        return None

    q = simpleQuery(aFile, "(set m where (in m memories) (in (select memport %s) (ports m)))" % getName(aPort))
    qlist = parseSet(q)
    if len(qlist) < 1:
        warning("Memport %s is not used in any memory." % str(aPort))
        return None

    if len(qlist) > 1:
        warning("Memport %s is used in more than one memory." % str(aPort))

    return qlist[0]

def getReadPorts(aFile, aMemory):
    if getType(aMemory) != TMEMORY:
        error("Expected memory, got %s." % str(aMemory))
        return None

    q = simpleQuery(aFile, "(set m where (in m (ports (select memory %s))) (read-port m))" % getName(aMemory))
    return parseSet(q)

def getWritePorts(aFile, aMemory):
    if getType(aMemory) != TMEMORY:
        error("Expected memory, got %s." % str(aMemory))
        return None

    q = simpleQuery(aFile, "(set m where (in m (ports (select memory %s))) (write-port m))" % getName(aMemory))
    return parseSet(q)

def getOutputs(aFile, aObj):
    q = simpleQuery(aFile, "(outputs (select %s %s))" % (getLongType(getType(aObj)), getName(aObj)))
    return parseSet(q)

def getInputs(aFile, aObj):
    q = simpleQuery(aFile, "(inputs (select %s %s))" % (getLongType(getType(aObj)), getName(aObj)))
    return parseSet(q)

def getDataout(aFile, aObj):
    q = simpleQuery(aFile, "(data-out (select %s %s))" % (getLongType(getType(aObj)), getName(aObj)))
    qlist = parseSet(q)

    if len(qlist) < 1:
        #warning("Object %s has no data-out." % str(aObj))
        return None

    if len(qlist) > 1:
        warning("Object %s has more than one data-out." % str(aObj))

    return qlist[0]

def getDatain(aFile, aObj):
    q = simpleQuery(aFile, "(data-in (select %s %s))" % (getLongType(getType(aObj)), getName(aObj)))
    qlist = parseSet(q)

    if len(qlist) < 1:
        #warning("Object %s has no data-in." % str(aObj))
        return None

    if len(qlist) > 1:
        warning("Object %s has more than one data-in." % str(aObj))

    return qlist[0]

def getBitwidth(aFile, aObj):
    q = simpleQuery(aFile, "(bitwidth (select %s %s))" % (getLongType(getType(aObj)), getName(aObj)))
    return parseInt(q)

def getAddress(aFile, aPort):
    if getType(aPort) != TMEM:
        error("Expected memory port, got %s." % str(aPort))
        return None

    q = simpleQuery(aFile, "(address (select %s %s))" % (getLongType(getType(aPort)), getName(aPort)))
    qlist = parseSet(q)

    if len(qlist) < 1:
        warning("Memory port %s has no address." % str(aPort))
        return None

    if len(qlist) > 1:
        warning("Memory port %s has more than one address." % str(aPort))

    return qlist[0]

def path(aFile, aSrc, aDst):
    q = simpleQuery(aFile, "(path (select %s %s) (select %s %s))" % \
        (getLongType(getType(aSrc)), getName(aSrc), getLongType(getType(aDst)), getName(aDst)))
    
    return q

def getPhaseFile(aFile, aObj):
    q = simpleQuery(aFile, "(getattr (select %s %s) \"ph\")" % \
        (getLongType(getType(aObj)), getName(aObj)))

    return parseInt(q)

def getWritePhases(aFile, aObj):
    q = simpleQuery(aFile, "(getattr (select %s %s) \"wphs\")" % \
        (getLongType(getType(aObj)), getName(aObj)))

    qlist = q.split(" ")
    phases = []
    for phase_str in qlist:
        phase = parseInt(phase_str)
        if phase == None:
            return []
        phases += [ phase ]

    return phases

def getReadPhases(aFile, aObj):
    q = simpleQuery(aFile, "(getattr (select %s %s) \"rphs\")" % \
        (getLongType(getType(aObj)), getName(aObj)))

    qlist = q.split(" ")
    phases = []
    for phase_str in qlist:
        phase = parseInt(phase_str)
        if phase == None:
            return []
        phases += [ phase ]

    return phases

def cacheAccessPhases(aFile, aStoragesOnly=False):
    wcache = {}
    rcache = {}
    pcache = {}
    if aStoragesOnly:
        objects = "registers memports"
    else:
        objects = "registers memports signals"
    q = simpleQuery(aFile, "(set o where (in o (union %s)))" % objects)
    qlist = parseSet(q)
    for obj in qlist:
        wcache[obj] = getWritePhases(aFile, obj)
        rcache[obj] = getReadPhases(aFile, obj)
        pcache[obj] = getPhaseFile(aFile, obj)
        
    return (wcache, rcache, pcache)

def getPhase(aPhases, aObj):
    return aPhases.get(aObj, None)

def getWrites(aWrites, aObj):
    return aWrites.get(aObj, None) 

def getReads(aReads, aObj):
    return aReads.get(aObj, None) 

def getPhaseDict(aPhases):
    '''
    Creates a mapping from each phase to its items.
    '''

    # Phase list
    phaseDict = {}
    # For each phase
    for phase in set(aPhases.values()):
        if phase != None:
            phaseDict[phase] = [ obj for (obj, ph) in aPhases.items() if ph == phase ]

    return phaseDict

def getReps(aFile, aPhaseDict):
    '''
    Gets a single representative pipeline register for each phase.
    '''
    # Dictionary of representatives
    reps = {}
    for (phase, storages) in aPhaseDict.items():
        pipe = [ i for i in storages if isPipe(aFile, getName(i), getType(i)) ]
        if len(pipe) > 0:
            reps[phase] = pipe[0]
        else:
            reps[phase] = None
            warning("No representative pipeline registers in phase %s" % str(phase))

    return reps

def getMicroArch(aFile):
    '''
    Returns a list of uarch. registers.
    '''
    # List containing each uarch. register
    micros = []

    q = simpleQuery(aFile, "(set r where (in r (union registers memports)))")
    qlist = parseSet(q)
    for obj in qlist:
        if isMicroArch(aFile, obj[0], obj[1]) and getPhaseFile(aFile, obj) != None:
            micros.append(obj)

    return micros

def getProgramCounter(aFile):
    q = simpleQuery(aFile, "(set r where (in r registers) (getattr r \"pc\"))")
    qlist = parseSet(q)

    if len(qlist) < 1:
        warning("No program counter found.")
        return None

    if len(qlist) > 1:
        warning("More than one program counter found. Returning first.")

    return qlist[0]

def parseSetArr(aQuery):
    if aQuery[:4] != "(set":
        return None

    arr = []
    par = 0
    current = ""
    for c in aQuery[5:-1]:
        if c == "(":
            par += 1

        if c == ")":
            par -= 1

        if par >= 1:
            if c.isspace():
                current += " "
            else:
                current = current + c
        elif c.isspace():
            if current != "":
                arr = arr + [current]
            current = ""
        elif par == 0:
            current = current + c

    if current != "":
        arr = arr + [current]

    return arr

def parseTupleArr(aQuery):
    if aQuery[:6] != "(tuple":
        return None

    arr = []
    par = 0
    current = ""
    for c in aQuery[7:-1]:
        if c == "(":
            par = par + 1

        if c == ")":
            par = par - 1

        if par >= 1:
            if c.isspace():
                current = current + " "
            else:
                current = current + c
        elif c.isspace():
            if current != "":
                arr = arr + [current]
            current = ""
        elif par == 0:
            current = current + c

    if current != "":
        arr = arr + [current]
 
    return tuple(arr)

def workListJoin(aWorkList, aItem):
    if not aItem in aWorkList:
        aWorkList = aWorkList.append(aItem)

def workListAppend(aWorkList, aItem):
    aWorkList = aWorkList.append(aItem)

def pause():
    '''
    Pauses execution.
    '''

    print "Press any key ..."
    raw_input()

# possible pre-defined filter options
REGISTER_FILTER = (lambda x,y,z: z==TREG)
PORT_FILTER     = (lambda x,y,z: z==TMEM)
PIPE_FILTER     = isPipe
MICRO_FILTER    = isMicroArch
ARCH_FILTER     = isArch
PIPE_MIC_FILTER = (lambda x,y,z: isPipe(x,y,z) or isMicroArch(x,y,z)) 
STORAGE_FILTER  = (lambda x,y,z: z==TREG or z==TMEM)

# default storage filter selects only registers
DEFAULT_STORAGE_FILTER=REGISTER_FILTER

def getStorages(aFile, aPhaseDict, aInclPhases=None, aExclPhases=None, \
    aFilterFunc=DEFAULT_STORAGE_FILTER):
    '''
    Returns registers that satisfies criteria given by filter function.

    @param  aFile       Input VAM file.
    @param  aPhaseDict  Dictionary mapping each phase to the set of its
                        elements (signals, registers, memports).
    @param  aInclPhases Included phases, defaults to all.
    @param  aExclPhases Excluded phases, defaults to the empty set.
    @param  aFilterFunc Filter function with interface
                        string(file) :: string(name) :: string(type) -> bool
                        (e.g. isPipe, isArch, isMicro)
    @return             A list of registers satisfying criteria.
    '''

    # Get all or selected phases
    if aInclPhases != None:
        phases = set(aInclPhases)
    else:
        phases = set(aPhaseDict.keys())

    # Remove excluded phases
    if aExclPhases != None:
        phases -= set(aExclPhases)

    # List of pipeline registers
    regs = []

    # For each accounted phase
    for phase in phases:
        regs += [ i for i in aPhaseDict[phase] if aFilterFunc(aFile, getName(i), getType(i)) ]

    return regs

def coiStorageSigs(aFile, aPhaseDict, aSig, aDir=COI_BCK, \
    aFilterFunc=DEFAULT_STORAGE_FILTER):
    '''
    Computes signals in front of registers that influence (are influenced by)
    the value of the given signal.

    @param  aFile       Input VAM file.
    @param  aPhaseDict  Dictionary mapping each phase to the set of its
                        elements (signals, registers, memports).
    @param  aSig        Signal for which the influencing signals are computed.
    @param  aDir        Direction of unroll.
    @param  aFilterFunc Filter function with an interface similar
                        to getStorages().
    @return             A set of output signals of registers.
    '''

    storages = getStorages(aFile, aPhaseDict, None, None, aFilterFunc)
    s_dir = ("out" if aDir==COI_BCK else "in")

    s_storage_sigs = ""
    for storage in storages:
        s_storage_sigs += "(data-%s (select %s %s)) " % \
            (s_dir, getLongType(getType(storage)), getName(storage))

    s_sig = "(select %s %s)" % (getLongType(getType(aSig)), getName(aSig))
    cmd = "(intersection (union (references (expr-%s 0 %s)) %s) (union %s))" % \
        (aDir, s_sig, s_sig, s_storage_sigs)
    q = simpleQuery(aFile, cmd)
    qlist = parseSet(q)

    return qlist

cs_memo = {}

def coiStorages(aFile, aPhaseDict, aSig, aDir=COI_BCK, \
    aFilterFunc=DEFAULT_STORAGE_FILTER):
    '''
    Computes registers that influence (are influenced by) the value
    of the given signal.

    @param  aFile       Input VAM file.
    @param  aPhaseDict  Dictionary mapping each phase to the set of its
                        elements (signals, registers, memports).
    @param  aSig        Signal for which the influencing signals are computed.
    @param  aDir        Direction of unroll.
    @param  aFilterFunc Filter function with an interface similar
                        to getStorages().
    @return             A set of output signals of registers.
    '''

    global cs_memo
    val_memo = cs_memo.get((aFile, aSig, aDir, aFilterFunc), None)
    if val_memo != None:
        return val_memo

    storages = getStorages(aFile, aPhaseDict, None, None, aFilterFunc)
    s_dir = ("out" if aDir==COI_BCK else "in")

    s_storages = ""
    for storage in storages:
        s_storages += "(select %s %s) " % \
            (getLongType(getType(storage)), getName(storage))

    s_sig = "(select %s %s)" % (getLongType(getType(aSig)), getName(aSig))
    cmd = "(set r | (in r (union %s)) \
               (intersection (data-%s r) \
                    (union (references (expr-%s 0 %s)) %s)))" % \
          (s_storages, s_dir, aDir, s_sig, s_sig)
    q = simpleQuery(aFile, cmd)
    qlist = parseSet(q)

    cs_memo[(aFile, aSig, aDir, aFilterFunc)] = qlist
    return qlist

def coiStorageNextSigs(aFile, aPhaseDict, aSig, aDir=COI_BCK, \
    aFilterFunc=DEFAULT_STORAGE_FILTER):
    '''
    Computes signals behind registers that influence (are influenced by)
    the value of the given signal.

    @param  aFile       Input VAM file.
    @param  aPhaseDict  Dictionary mapping each phase to the set of its
                        elements (signals, registers, memports).
    @param  aSig        Signal for which the influencing signals are computed.
    @param  aDir        Direction of unroll.
    @param  aFilterFunc Filter function with an interface similar
                        to getStorages().
    @return             A set of output signals of registers.
    '''

    storages = coiStorages(aFile, aPhaseDict, aSig, aDir, aFilterFunc)

    s_dir = ("in" if aDir==COI_BCK else "out")
    s_storage_sigs = ""
    for storage in storages:
        s_storage_sigs += "(data-%s (select %s %s)) " % \
            (s_dir, getLongType(getType(storage)), getName(storage))

    cmd = "(union %s)" % s_storage_sigs
    q = simpleQuery(aFile, cmd)
    qlist = parseSet(q)

    return qlist

def unrollExprs():
    pass

usigs_memo = {}

def unrollSigs(aFile, aPhaseDict, aSigs, aK=0, aDir=COI_BCK, aOrigin=0):
    global usigs_memo

    # Currently processed signals
    sigs = frozenset(aSigs)
    val_memo = usigs_memo.get((aFile, sigs, aK, aDir, aOrigin), None)
    if val_memo != None:
        return val_memo

    # Dictionary T -> 2^Sigs, i.e. its items pair time with signals
    unroll = {}
    # Perform unroll
    i = 0
    # Current relative time
    t = aOrigin
    # Update unroll dictionary
    unroll.update({ t : sigs })
    while i < aK:
        # Compute new influencing signals
        new_sigs = set([])
        for sig in sigs:
            coi_sigs = coiStorageNextSigs(aFile, aPhaseDict, sig, aDir, \
                PIPE_MIC_FILTER)
            new_sigs |= set(coi_sigs)
        sigs = frozenset(new_sigs)

        # Proceed to the next phase
        if aDir == COI_BCK: t -= 1
        else:               t += 1
       
        # Update unroll dictionary
        unroll.update({ t : sigs })

        # Increment step counter
        i += 1

    usigs_memo[(aFile, sigs, aK, aDir, aOrigin)] = unroll
    return unroll

def unrollMerge(*unroll_args):
    merged = {}
    for unroll in unroll_args:
        for phase in unroll:
            if not phase in merged:
                merged[phase] = unroll[phase]
            else:
                merged[phase] |= unroll[phase]

    return merged

def unrollShift(aUnroll, i):
    unroll = {}
    for phase in aUnroll:
        unroll[phase+i] = aUnroll[phase]

    return unroll

def applyPhaseConds(aFile, aPhaseDict, aCtrlSigs, aPath, aQ, \
    aLabel, aDir=COI_BCK):

    # Construct a sequence of conditions that hold along the given path
    conds = []
    i = 0
    for q in aPath:
        if q == aQ:
            # Remember index of state aQ
            idx = i
        conds.append(aLabel[q])
        i += 1

    return applyPhaseCondsImpl(aFile, aPhaseDict, aCtrlSigs, conds,
        idx, aDir)

def applyPhaseCondsImpl(aFile, aPhaseDict, aCtrlSigs, aConds, \
    aIdx, aDir=COI_BCK):

    # initializations
    sz = len(aConds)
    u = set([])

    # for each set of conditions in the list
    for i in range(sz):
        # take only object part of conditions (i.e., without the actual
        # constrained value)
        conds = [ c.obj() for c in aConds[i] ]
        if aIdx == i:
            # special condition -- control sigs are assumed to hold
            # in the stage given by input index
            conds += aCtrlSigs

        # distance from the input index
        dist = i - aIdx
        if aDir == COI_BCK:
            # unroll backwards for i steps with origin given by the distance
            # from the input index
            unew = unrollSigs(aFile, aPhaseDict, conds, i, COI_BCK, dist)
        else:
            # inversion of index for forward case
            ifwd = sz-i-1
            # unroll forwards for ifwd steps with origin given by the distance
            # from the input index
            unew = unrollSigs(aFile, aPhaseDict, conds, ifwd, COI_FWD, dist)

        # merge unrolls over the already computed ones
        u = unrollMerge(u, unew)

    return u

# TODO: Refactor ...
def unrollConds(aFile, aPhaseDict, aConds, aIdx, aDir=COI_BCK):
    sz = len(aConds)
    u = set([])
    for n_steps_bck in range(sz):
        n_steps_fwd = sz - n_steps_bck - 1
        dist = n_steps_bck - aIdx
        conds = [ c.obj() for c in aConds[n_steps_bck] ]
        if aDir == COI_BCK:
            unew = unrollSigs(aFile, aPhaseDict, conds, n_steps_bck, COI_BCK, dist)
        else:
            unew = unrollSigs(aFile, aPhaseDict, conds, n_steps_fwd, COI_FWD, dist)
        u = unrollMerge(u, unew)

    return u

#def unrollPath(aFile, aPhaseDict, aPath, aQ, Label):
#    # Construct a sequence of conditions that hold along the path
#    conds = []
#    i = 0
#    for q in path:
#        if q == aQ:
#            idx = i
#        conds.append(Label[q])
#        i += 1
#
#    return unrollConds(aFile, aPhaseDict, conds, idx)
#
#def condSat(aFile, aPhaseDict, aCond, aPath1, Label1, aUr1, aPath2, aUr2, Label2, aDefs, aUAsserts):
#
#    # Get definition of relevant signals
#    # At first, collect all uarch regs
#    mclause = ""
#    for micro in aUarchs:
#        mclause = mclause + "(data-out (select register %s)) " % str(micro[0])
#
#    return ""

def upc2smt(aFile, aPhaseDict, upcs, aAuxSigDecls=[], aAuxSigAsserts=""):
    cmd_decls = ""
    cmd_coi = ""
    cmd_regs = ""
    s_conds = ""
    out = ""

    # Pre-compute some values
    primes = set([])
    tagged_unroll = {}
    for upc in upcs:
        # Collect all possible primings
        primes |= set(upc.unroll().keys())
        # Tag unrolls using the parent's tag
        for prime in upc.unroll():
            tagged_unroll[prime] = tagged_unroll.get(prime, frozenset([])) | \
                frozenset([ (upc.dir(), upc.tag(), upc.unroll()[prime]) ])

    primes = sorted(primes)

    # Get declarations for each unrolled conditions
    #print "cmd_decls", "cmd_coi"
    for prime in primes:
        # A zero priming is a special case 
        if prime == 0:
            # Declarations are blended into each other
            cmd_uni_decls = "(union "
            cmd_uni_coi   = "(union "
            # Include auxiliary signals first
            for sig in sorted(aAuxSigDecls):
                cmd_uni_decls += "(select signal %s) " % getName(sig)

            # Then, proceed to to signals and expressions
            # of the current phase
            for (dir_, tag, sigs) in sorted(tagged_unroll[prime]):
                s_coi_dir = "coib" if dir_==COI_BCK else "coi"
                for sig in sorted(sigs):
                    coi_clause = "(expr-%s 0 (select signal %s)) " % \
                        (s_coi_dir, getName(sig))
                    cmd_uni_decls += "(references " + coi_clause + ") " \
                        "(select signal %s) " % getName(sig)
                    cmd_uni_coi += coi_clause
            cmd_uni_decls += ") "
            cmd_uni_coi   += ") "

            # Create (smtlib) query
            cmd_decls += "(smtlib-tag-prime \"\" 0 " + cmd_uni_decls + ") "
            cmd_coi   += "(smtlib-tag-prime \"\" 0 " + cmd_uni_coi   + ") "
        # A non-zero case
        else:
            # Declarations are kept separated (using unique tag)
            for (dir_, tag, sigs) in sorted(tagged_unroll[prime]):
                s_coi_dir = "coib" if dir_==COI_BCK else "coi"

                cmd_uni_decls = "(union "
                cmd_uni_coi   = "(union "
                # Then, proceed to to signals and expressions
                # of the current phase
                for sig in sorted(sigs):
                    coi_clause = "(expr-%s 0 (select signal %s)) " % \
                        (s_coi_dir, getName(sig))
                    cmd_uni_decls += "(references " + coi_clause + ") " \
                        "(select signal %s) " % getName(sig)
                    cmd_uni_coi += coi_clause
                cmd_uni_decls += ") "
                cmd_uni_coi   += ") "

                # Create (smtlib) query
                cmd_decls += "(smtlib-tag-prime \"%s\" %d %s) " % \
                    (tag, prime, cmd_uni_decls)
                cmd_coi   += "(smtlib-tag-prime \"%s\" %d %s) " % \
                    (tag, prime, cmd_uni_coi)

    #print "coi", cmd_coi
    #print cmd_decls
    q = simpleQuery(aFile, cmd_decls)
    out += parseSmtlibExport(q, "Declarations")

    # Declarations from symbolic conditions
    decls = set([])
    for upc in upcs:
        for (prime, conds) in upc.phaseConds().items():
            for cond in conds:
                decls |= set(cond.decls())

    for decl in decls:
        out += decl + "\n\n"

    # Get cone of influence for each unrolled conditions
    #print "cmd_coi"
    #for prime in primes:
    #    # A zero priming is a special case 
    #    if prime == 0:
    #        # Expressions are united
    #        cmd_uni = "(union "
    #        for (dir_, tag, sigs) in tagged_unroll[prime]:
    #            s_coi_dir = "coib" if dir_==COI_BCK else "coi"
    #            for sig in sigs:
    #                cmd_uni += "(expr-%s 0 (select signal %s)) " % \
    #                    (s_coi_dir, getName(sig))
    #        cmd_uni += ") "
    #        cmd_coi += "(smtlib-tag-prime \"\" 0 %s) " % \
    #            cmd_uni
    #    # A non-zero case
    #    else:
    #        # Expressions are kept separated (using unique tag)
    #        for (dir_, tag, sigs) in tagged_unroll[prime]:
    #            cmd_uni = "(union "
    #            s_coi_dir = "coib" if dir_==COI_BCK else "coi"
    #            for sig in sigs:
    #                cmd_uni += "(expr-%s 0 (select signal %s)) " % \
    #                    (s_coi_dir, getName(sig))
    #            cmd_uni += ") "
    #            cmd_coi += "(smtlib-tag-prime \"%s\" %d %s) " % \
    #                (tag, prime, cmd_uni)
 
    #print cmd_coi
    q = simpleQuery(aFile, cmd_coi)
    out += parseSmtlibExport(q, "Cone of influence") + "\n"

    # Get intermediate registers for each unrolled conditions
    #print "cmd_regs"
    for upc in upcs:
        for prime in upc.unroll():
            # Compute next phase
            n = (prime-1 if upc.dir()==COI_BCK else prime)
            if not n in upc.unroll():
                continue

            regs = set([])
            for sig in upc.unroll()[prime]:
                # Collect all registers influenced by signals of unroll
                regs |= set(coiStorages(aFile, aPhaseDict, sig, upc.dir(),
                    PIPE_MIC_FILTER))

            # Does the current cone of influence include any registers?
            if len(regs) == 0:
                continue

            # If yes, capture them all
            cmd_uni1 = "(union "
            for reg in regs:
                cmd_uni1 += "(select register %s) " % getName(reg)
            cmd_uni1 += ") "

            # Signals of registers of the other phase
            cmd_uni2  = "(union "
            for sig in upc.unroll()[n]:
                cmd_uni2 += "(select signal %s) " % getName(sig)
            cmd_uni2 += ") "

            # Find the intersecting registers
            s_dir = ("in" if upc.dir()==COI_BCK else "out")
            cmd_set = "(set r | (in r %s) " \
                "(intersection (data-%s r) %s))" % \
                (cmd_uni1, s_dir, cmd_uni2)

            # Extract the registers
            cmd_regs += "(smtlib-tag-prime \"%s\" %d %s) " % \
                (upc.tag(), n, cmd_set)

    #print cmd_regs
    q = simpleQuery(aFile, cmd_regs)
    out += parseSmtlibExport(q, "Registers") + "\n"

    #print "cmd_conds"
    for upc in upcs:
        for (prime, conds) in upc.phaseConds().items():
            for cond in conds:
                s_conds += cond.smt(upc.tag(), prime) + "\n"

    if s_conds != "":
        out += ";;; Phase conditions\n"
        out += s_conds + "\n"

    if aAuxSigAsserts != "":
        out += ";;; Additional auxiliary assertions\n"
        out += aAuxSigAsserts + "\n"
 
    return out

#            next_prime = (prime+1 if upc.dir()==COI_BCK else prime-1)
#            if next_prime in upc.unroll():
#                # Compute assertions describing transition over registers
#                s_dir2 = ("out" if upc.dir()==COI_BCK else "in" )
#                s_dir1 = ("in"  if upc.dir()==COI_BCK else "out")
#
#                # Signals of registers of the current phase
#                cmd_uni1  = "(union "
#                for sig in upc.unroll()[prime]:
#                    cmd_uni1 += "(select signal %s) " % getName(sig)
#                cmd_uni1 += ") "
#
#                # Signals of registers of the other phase
#                cmd_uni2  = "(union "
#                for sig in upc.unroll()[next_prime]:
#                    cmd_uni2 += "(select signal %s) " % getName(sig)
#                cmd_uni2 += ") "
#
#                # Find the intersecting registers
#                cmd_set = "(set r | (in r registers) " \
#                    "(intersection (data-%s r) %s) " \
#                    "(intersection (data-%s r) %s))" % \
#                    (s_dir1, cmd_uni1, s_dir2, cmd_uni2)
#                                        
#                # Extract the registers
#                n = (prime+1 if upc.dir()==COI_BCK else prime)
#                cmd_regs += "(smtlib-prime %d %s) " % \
#                    (n, cmd_set)

#        # Get declarations 
#        cmd_decls = "(smtlib-tag-prime %s %d (union " % \
#            (upc.tag(), prime)
#
#        s_coi_dir = "coib" if upc.dir()==COI_BCK else "coi"
#        primes = sorted(upc.unroll().keys())
#        for prime in primes:
#            # Get declarations 
#            cmd_decls = "(smtlib-tag-prime %s %d (union " % \
#                (upc.tag(), prime)
#            for sig in upc.unroll()[prime]:
#                cmd_decls += "(references " \
#                    "(expr-%s 0 (select signal %s))) " \
#                    "(select signal %s) " % \
#                    (s_coi_dir, getName(sig), getName(sig))
#            cmd_decls += ")) "
#            print cmd_decls
#
#            # Extract cone of influence for each signal
#            cmd_coi = "(smtlib-tag-prime %s %d (union " % \
#                (upc.tag(), prime)
#            for sig in upc.unroll()[prime]:
#                cmd_coi += "(expr-%s 0 (select signal %s)) " % \
#                    (s_coi_dir, getName(sig))
#            cmd_coi += ")) "
#            print cmd_coi
#
#            next_prime = (prime+1 if upc.dir()==COI_BCK else prime-1)
#            if next_prime in upc.unroll():
#                # Compute assertions describing transition over registers
#                s_dir1 = ("out" if upc.dir()==COI_BCK else "in" )
#                s_dir2 = ("in"  if upc.dir()==COI_FWD else "out")
#
#                # Signals of registers of the current phase
#                cmd_uni1  = "(union "
#                for sig in upc.unroll()[prime]:
#                    cmd_uni1 += "(select signal %s) " % getName(sig)
#                cmd_uni1 += ") "
#
#                # Signals of registers of the other phase
#                cmd_uni2  = "(union "
#                for sig in upc.unroll()[next_prime]:
#                    cmd_uni2 += "(select signal %s) " % getName(sig)
#                cmd_uni2 += ") "
#
#                # Find the intersecting registers
#                cmd_regs = "(set r | (in r registers) " \
#                    "(intersection (data-%s r) %s) " \
#                    "(intersection (data-%s r) %s))" % \
#                    (s_dir1, cmd_uni1, s_dir2, cmd_uni2)
#                                        
#                # Extract the registers
#                n = (prime+1 if upc.dir()==COI_BCK else prime)
#                cmd_regs = "(smtlib-tag-prime %s %d %s)" % \
#                    (upc.tag(), n, cmd_regs)
#                print cmd_regs

class UnrolledPhaseConds():
    def __init__(self, aPhaseConds, aUnroll, aDir=COI_BCK, aTag=""):
        self._phase_conds = aPhaseConds
        self._unroll = aUnroll
        self._dir = aDir
        self._tag = aTag

    @classmethod
    def fromfile(cls, aFile, aPhaseDict, aCtrlSigs, aPath, aQ, \
        aLabel, aDir=COI_BCK, aTag=""):

        # Extract index of aQ in path aPath
        conds = []
        i = 0
        for q in aPath:
            if q == aQ:
                idx = i
            # Construct list of phase conditions as well
            conds.append(aLabel[q])
            i += 1

        # Construct a sequence of phase conditions shifted
        # by the computed index
        phase_conds = {}
        i = -idx
        for cond in conds:
            phase_conds[i] = cond
            i += 1

        unroll = applyPhaseConds(aFile, aPhaseDict, aCtrlSigs,
            aPath, aQ, aLabel, aDir)

        return cls(phase_conds, unroll, aDir, aTag)

    def shiftOrigin(self, aShift):
        new_unroll = { phase+aShift:val for (phase,val) \
            in self._unroll.items() }
        new_conds = { phase+aShift:val for (phase,val) \
            in self._phase_conds.items() }

        return UnrolledPhaseConds(new_conds, new_unroll,
            self.dir(), self.tag())

    def merge(self, aOther):
        if self.tag() != aOther.tag() or self.dir() != aOther.dir():
            return None

        new_unroll = unrollMerge(self.unroll(), aOther.unroll())
        new_conds = {}
        for (phase,conds) in self.phaseConds().items() + aOther.phaseConds().items():
            val = list(set(new_conds.get(phase, []) + conds))
            new_conds[phase] = val

        return UnrolledPhaseConds(new_conds, new_unroll,
            self.dir(), self.tag())

    def phaseConds(self):
        return self._phase_conds

    def unroll(self):
        return self._unroll

    def dir(self):
        return self._dir

    def tag(self):
        return self._tag
        
class Cond():
    def __init__(self, aObj, aBw, aComm=""):
        self._obj   = aObj
        self._bw    = int(aBw)
        self._comm  = "" if aComm=="" else " ; " + aComm

    def obj(self):
        return self._obj

    def name(self):
        return getName(self._obj)

    def type(self):
        return getType(self._obj)

    def bw(self):
        return self._bw

    def decls(self):
        return []

    def comm(self):
        return self._comm

    def val(self):
        error("Method not implemented!")
        return None

class SymCond(Cond):
    def __init__(self, aObj, aSym, aBw, aComm=""):
        Cond.__init__(self, aObj, aBw, aComm)
        self._sym = aSym

    @classmethod
    def fromstring(cls, aObj, aStr, aBw):
        aStr = aStr.strip()

        match = re.match("[a-zA-Z_][0-9a-zA-Z_-]*", aStr)
        if match != None:
            return cls(aObj, aStr, aBw)

        error("Error during construction of constant condition "
            "for object '%s' with value '%s'" % (str(aObj), str(aStr)))
        return None

    # override
    def decls(self):
        return [ "(declare-fun %s () (_ BitVec %d))" % (self.val(), self.bw()) ]

    def val(self):
        return self._sym

    def smt(self, aTag="", aPrime=0):
        if   aPrime > 0:    prime = "_{!s}p".format( aPrime)
        elif aPrime < 0:    prime = "_{!s}m".format(-aPrime)
        else:               prime = ""

        if aPrime != 0 and aTag != "":  tag = "_%s" % aTag
        else:                           tag = ""

        return "(assert (= %s%s%s %s))%s" % (self.name(), tag,
            prime, self.val(), self.comm())

    def __eq__(self, aRhs):
        if not isinstance(aRhs, SymCond):
            return False

        return self.obj() == aRhs.obj() and self.val() == aRhs.val() and self.bw() == aRhs.bw()

    def __neq__(self, aRhs):
        return not (self == aRhs)

    def __repr__(self):
        return "(= %s %s))" % (self.name(), self.val())

    def __hash__(self):
        return hash((self._obj, self._sym, self._bw))

class ConstCond(Cond):
    def __init__(self, aObj, aVal, aBw, aComm=""):
        Cond.__init__(self, aObj, aBw, aComm)
        self._val = int(aVal)

    @classmethod
    def fromstring(cls, aObj, aStr):
        aStr = aStr.strip()
        match = re.match("\(\s*_\s+bv(\d+)\s+(\d+)\s*\)", aStr)
        if match != None:
            return cls(aObj, int(match.group(1)), int(match.group(2)))

        match = re.match("\(\s*_\s+(\d+)\s+(\d+)\s*\)", aStr)
        if match != None:
            return cls(aObj, int(match.group(1)), int(match.group(2)))
    
        match = re.match("#b([01]+)", aStr)
        if match != None:
            return cls(aObj, int(match.group(1), 2), len(match.group(1)))

        match = re.match("#x([0-9a-fA-F]+)", aStr)
        if match != None:
            return cls(aObj, int(match.group(1), 16), len(match.group(1))*4)

        error("Error during construction of constant condition "
            "for object '%s' with value '%s'" % (str(aObj), str(aStr)))
        return None

    def val(self):
        return self._val

    def smt(self, aTag="", aPrime=0, aNegate=False):
        if   aPrime > 0:    prime = "_{!s}p".format( aPrime)
        elif aPrime < 0:    prime = "_{!s}m".format(-aPrime)
        else:               prime = ""

        if aPrime != 0 and aTag != "":  tag = "_%s" % aTag
        else:                           tag = ""

        if aNegate:
            ret = "(assert (not (= %s%s%s (_ bv%d %d))))%s" % (self.name(),
                tag, prime, self.val(), self.bw(), self.comm())
        else:
            ret = "(assert (= %s%s%s (_ bv%d %d)))%s" % (self.name(), tag,
                prime, self.val(), self.bw(), self.comm())

        return ret

    def __eq__(self, aRhs):
        if not isinstance(aRhs, ConstCond):
            return False

        return self.obj() == aRhs.obj() and self.val() == aRhs.val() and self.bw() == aRhs.bw()

    def __neq__(self, aRhs):
        return not (self == aRhs)

    def __repr__(self):
        return "(= %s (_ bv%d %d))" % (self.name(), self.val(), self.bw())

    def __hash__(self):
        return hash((self._obj, self._val, self._bw))

def getTimeStatistics():
    global g_rtlmanip_time
    global g_smt_time
    global g_armc_time
 
    return (g_rtlmanip_time, g_smt_time, g_armc_time)

#def isCondSAT(aFile, aPhaseDict, aConds1, aConds2, aIdx1, aIdx2, aCond):
    

## We can unroll a whole set of signals, however these signals
## must originate from the same stage given by aFrom argument.
## TODO: Maybe the following check is too much expensive ...
#for sig in aSigs:
#    if aFrom != getPhaseFile(aFile, sig):
#        # Error
#        error("Signal '%s' is not from phase %d" % \
#            (getName(sig), (aFrom-1)))
#        return None

## Determine filter function for COI computation
#if aDir == COI_BCK:
#    # Accepts signals of pipeline registers that are written
#    # by the previous phase
#    filter_func = lambda file_, name, type_: \
#        isPipe(file_, name, type_) and \
#        getPhaseFile(file_, (name, type_)) == phase
#else:
#    # Accepts signals of pipeline registers that are read
#    # by the next phase
#    filter_func = lambda file_, name, type_: \
#        isPipe(file_, name, type_) and \
#        getPhaseFile(file_, (name, type_)) == (phase+1)

#def findPaths(Q, Delta, qin):
#    # TODO: Implement ...
#    return [ Q ]
#
#def unrollState(aFile, aPhaseDict, aPath, aQ, Label):
#    (Q, Delta, qin, Label) = aTS
#    paths = findPaths(Q, Delta, qin)
#    # TODO: Implement ...
#    if len(paths) != 1:
#        error("Unroll is not implement for the number of paths that differs from 1.")
#        return None
#
#    # Construct a separate unroll for each found path
#    unrolls = []
#    for path in paths

