#!/usr/bin/python
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Name:         consistency.py
# Description:  Performs consistency check of a cpu pipeline.
# Author:       Lukas Charvat 25.2.2014
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import sys
import os
from common import *

debug_con = False
debug_stop = False
debug_dots = True

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Consistency checking functions
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def checkStall(aFile, aPhaseDict, aReps, aMicros, aAsserts):
    '''
    (A) Check stall
    ~~~~~~~~~~~~~~~
    If there exists a stalled register, or register with WE set to 0
    in phase X, then:
    (1) all registers in phase X must:
      (1.1) be stalled, or
      (1.2) have WE set to 0.
    (2) all registers in phases Y < X must:
      (2.1) be stalled, or
      (2.2) be cleared, or
      (2.3) have WE set to 0.
    (3) all registers in phase Z = X + 1 must:
      (3.1) be stalled, or
      (3.2) be cleared, or
      (3.3) have WE set to 0.
    '''

    supress_err = True

    # Check part described in (A.1)
    # Proof of s_1 <=> ... <=> s_k using series of implications of the form:
    #   s_1 => s_2 => ... => s_k => s_1
    info("Checking consistency part (A.1) ...")
    for phase in aPhaseDict:
        pipe = [ i for i in aPhaseDict[phase] if isPipe(aFile, i[0], i[1]) ]
        for i in range(len(pipe)):
            # Index of next storage (modulo by the no. of storages)
            j = (i + 1) % len(pipe)

            # Get all required expression for analysis
            cmd_core = "(union (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s))) (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s))))" % \
            (getLongType(pipe[i][1]), pipe[i][0], \
             getLongType(pipe[i][1]), pipe[i][0], \
             getLongType(pipe[j][1]), pipe[j][0], \
             getLongType(pipe[j][1]), pipe[j][0])

            # Header for file
            smtlibstr = ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n" 
            smtlibstr += ";;; SMTTOOL run for phase %s (%s => %s)\n" % \
                (pipe, pipe[i][0], pipe[j][0])

            # Get definition of relevant signals
            # At first, collect all uarch regs
            mclause = ""
            for micro in aMicros:
                mclause += "(data-out (select register %s)) " % str(micro[0])
            # Then, merge them with previously extracted expression
            q = simpleQuery(aFile, "(smtlib (union (references %s) %s))" % (cmd_core, mclause))
            smtlibstr += parseSmtlibExport(q) + "\n"

            # Get expresions that holds for given signals
            q = simpleQuery(aFile, "(smtlib %s)" % cmd_core)
            smtlibstr += parseSmtlibExport(q) + "\n"

            # Extract stall and enable signal of current storage
            st = getCtrl(aFile, pipe[i], CST)
            we = getCtrl(aFile, pipe[i], CEN)

            # Build assertions for SMTLib
            if st != None:
                smtlibstr += smtlibExport("(assert (or (= %s #b1) (= %s #b0)) )\n" % (st[0], we[0]))
            else:
                smtlibstr += smtlibExport("(assert (= %s #b0) )\n" % we[0])

            # Extract stall and enable signal of next storage
            st = getCtrl(aFile, pipe[j], CST)
            we = getCtrl(aFile, pipe[j], CEN)

            # Build assertions for SMTLib
            if st != None:
                smtlibstr += smtlibExport("(assert (and (= %s #b0) (= %s #b1)) )\n" % (st[0], we[0]))
            else:
                smtlibstr += smtlibExport("(assert (= %s #b1) )\n" % we[0])

            # Additional assertions based on observation of state registers
            smtlibstr += aAsserts

            # Add addtional commands for SMTLib
            smtlibstr += smtlibExport("(check-sat)\n(exit)") + "\n"

            # Run SMTTOOL
            if debug_con:
                info("SMTTOOL run for (%s => %s)" % (pipe[i][0], pipe[j][0]))
            run = runsmt(smtlibstr)
            if run != SMTUNSAT:
                error("Stalling of pipeline register '%s' (%s) in phase %d does not "
                      "imply stalling of pipeline register '%s' (%s) in the same phase!" % \
                    (getName(pipe[i]), getComment(aFile, getName(pipe[i]), getType(pipe[i])), phase, \
                     getName(pipe[j]), getComment(aFile, getName(pipe[j]), getType(pipe[j]))))
                error("SMTTOOL run for (%s => %s)" % (getName(pipe[i]), getName(pipe[j])))
                error(smtrun2str(run))
                if not supress_err:
                    supress_err = questionYesNo("Supress further consistency "
                        "errors? (Y/n)", True)
            else:
                if debug_con:
                    info("SMTTOOL Result: %s" % smtrun2str(run))
                else:
                    if debug_dots:
                        sys.stdout.write(".")
                        sys.stdout.flush()

    if not debug_con and debug_dots:
        print

    info("Checking consistency part (A.2) ...")
    # Check part described in (A.2). Because of (A.1) and (B.1), it is enough
    # to take one representative from each phase.

    # Checking of (A.2) itself.
    for r in aReps:
        # Representative from phase X.
        rep = aReps[r]
        if rep == None:
            continue

        for r1 in [ i for i in aReps if i < r ]:
            # Representative from phase Y.
            rep1 = aReps[r1]
            if rep1 == None:
                continue

            # Get all required expression for analysis
            cmd_core = "(union (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s)))(expr-coib 0 (clear (select %s %s))) (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s))) (expr-coib 0 (clear (select %s %s))))" % \
            (getLongType(rep[1]), rep[0], \
             getLongType(rep[1]), rep[0], \
             getLongType(rep[1]), rep[0], \
             getLongType(rep1[1]), rep1[0], \
             getLongType(rep1[1]), rep1[0], \
             getLongType(rep1[1]), rep1[0])

            # Header for file
            smtlibstr = ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n" 
            smtlibstr += ";;; SMTTOOL run for phase (st or we %s =>- st or we or cl %s)\n" % \
                (rep[0], rep1[0])

            # Get definition of relevant signals
            # At first, collect all uarch regs
            mclause = ""
            for micro in aMicros:
                mclause += "(data-out (select register %s)) " % str(micro[0])
            # Then, merge them with previously extracted expression
            q = simpleQuery(aFile, "(smtlib (union (references %s) %s))" % (cmd_core, mclause))
            smtlibstr += parseSmtlibExport(q) + "\n"

            # Get expresions that holds for given signals
            q = simpleQuery(aFile, "(smtlib %s)" % cmd_core)
            smtlibstr += parseSmtlibExport(q) + "\n"

            # Extract stall and enable signal of representative from phase X.
            st = getCtrl(aFile, rep, CST)
            we = getCtrl(aFile, rep, CEN)
            cl = getCtrl(aFile, rep, CCL)

            # Build assertions for SMTLib
            if st != None:
                smtlibstr += smtlibExport("(assert (or (= %s #b1) (and (= %s #b0) (= %s #b0))) ) ; st we cl\n" % (st[0], we[0], cl[0]))
            else:
                smtlibstr += smtlibExport("(assert (= %s #b0) ) ; we\n" % we[0])

            # Extract stall, enable and clear signal of representative from phase Y.
            st = getCtrl(aFile, rep1, CST)
            we = getCtrl(aFile, rep1, CEN)
            cl = getCtrl(aFile, rep1, CCL)

            # Build assertions for SMTLib
            if st != None and cl != None:
                smtlibstr += smtlibExport("(assert (and (= %s #b0) (and (= %s #b1) (= %s #b0))) ) ; st we cl\n" % (st[0], we[0], cl[0]))
            else:
                smtlibstr += smtlibExport("(assert (= %s #b1) ) ; we\n" % we[0])

            # Additional assertions based on observation of state registers
            smtlibstr += aAsserts

            # Add addtional commands for SMTLib
            smtlibstr += smtlibExport("(check-sat)\n(exit)") + "\n"

            # Run SMTTOOL
            if debug_con:
                info("SMTTOOL run for (%s => %s)" % (rep[0], rep1[0]))
            run = runsmt(smtlibstr)
            if run != SMTUNSAT:
                error("Stalling of pipeline register '%s' (%s) in phase %d does not "
                      "imply stalling of pipeline register '%s' (%s) in phase %d!" % \
                    (getName(rep),  getComment(aFile, getName(rep), getType(rep)),   r, \
                     getName(rep1), getComment(aFile, getName(rep1), getType(rep1)), r1))
                error("SMTTOOL run for (%s => %s)" % (getName(rep), getName(rep1)))
                error(smtrun2str(run))
                if not supress_err:
                    supress_err = questionYesNo("Supress further consistency "
                        "errors? (Y/n)", True)
            else:
                if debug_con:
                    info("SMTTOOL Result: %s" % smtrun2str(run))
                else:
                    if debug_dots:
                        sys.stdout.write(".")
                        sys.stdout.flush()

    if not debug_con and debug_dots:
        print

    # Check of part (A.3)
    info("Checking consistency part (A.3) ...")
    for r in aReps:
        # Representative from phase X.
        rep = aReps[r]
        if rep == None:
            continue

        # Representative from phase Z.
        rep1 = aReps.get(r + 1, None)
        if rep1 == None:
            continue

        # Get all required expression for analysis
        cmd_core = "(union (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s))) (expr-coib 0 (clear (select %s %s))) (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s))) (expr-coib 0 (clear (select %s %s))))" % \
        (getLongType(rep[1]), rep[0], \
         getLongType(rep[1]), rep[0], \
         getLongType(rep[1]), rep[0], \
         getLongType(rep1[1]), rep1[0], \
         getLongType(rep1[1]), rep1[0], \
         getLongType(rep1[1]), rep1[0])

        # Header for file
        smtlibstr = ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n" 
        smtlibstr += ";;; SMTTOOL run for phase (st or we %s =>+ st or we or cl %s)\n" % \
            (rep[0], rep1[0])

        # Get definition of relevant signals
        # At first, collect all uarch regs
        mclause = ""
        for micro in aMicros:
            mclause += "(data-out (select register %s)) " % str(micro[0])
        # Then, merge them with previously extracted expression
        q = simpleQuery(aFile, "(smtlib (union (references %s) %s))" % (cmd_core, mclause))
        smtlibstr += parseSmtlibExport(q) + "\n"

        # Get expresions that holds for given signals
        q = simpleQuery(aFile, "(smtlib %s)" % cmd_core)
        smtlibstr += parseSmtlibExport(q) + "\n"

        # Extract stall and enable signal of representative from phase X.
        st = getCtrl(aFile, rep, CST)
        we = getCtrl(aFile, rep, CEN)
        cl = getCtrl(aFile, rep, CCL)

        # Build assertions for SMTLib
        if st != None:
            smtlibstr += smtlibExport("(assert (or (= %s #b1) (and (= %s #b0) (= %s #b0))) ) ; st we cl\n" % (st[0], we[0], cl[0]))
        else:
            smtlibstr += smtlibExport("(assert (= %s #b0) )\n" % we[0])

        # Extract stall, enable and clear signal of representative from phase Z.
        st = getCtrl(aFile, rep1, CST)
        we = getCtrl(aFile, rep1, CEN)
        cl = getCtrl(aFile, rep1, CCL)

        # Build assertions for SMTLib
        if st != None and cl != None:
            smtlibstr += smtlibExport("(assert (and (= %s #b0) (and (= %s #b1) (= %s #b0))) ) ; st we cl\n" % (st[0], we[0], cl[0]))
        else:
            smtlibstr += smtlibExport("(assert (= %s #b1) )\n" % we[0])

        # Additional assertions based on observation of state registers
        smtlibstr += aAsserts

        # Add addtional commands for SMTLib
        smtlibstr += smtlibExport("(check-sat)\n(exit)") + "\n"

        # Run SMTTOOL
        if debug_con:
            info("SMTTOOL run for (%s => %s)" % (rep[0], rep1[0]))
        run = runsmt(smtlibstr)
        if run != SMTUNSAT:
            error("Stalling of pipeline register '%s' (%s) in phase %d does not "
                  "imply clearing of pipeline register '%s' (%s) in phase %d!" % \
                (getName(rep),  getComment(aFile, getName(rep), getType(rep)),   r, \
                 getName(rep1), getComment(aFile, getName(rep1), getType(rep1)), r + 1))
            error("SMTTOOL run for (%s => %s)" % (getName(rep), getName(rep1)))
            error(smtrun2str(run))
            if not supress_err:
                supress_err = questionYesNo("Supress further consistency "
                    "errors? (Y/n)", True)
        else:
            if debug_con:
                info("SMTTOOL Result: %s" % smtrun2str(run))
            else:
                if debug_dots:
                    sys.stdout.write(".")
                    sys.stdout.flush()

    if not debug_con and debug_dots:
        print

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def checkClear(aFile, aPhaseDict, aReps, aMicros, aAsserts):
    '''
    If there exists a cleared register in phase X, then:
    (1) all registers in phase X must be cleared,
    (2) (FIXME: too strong) all registers in phase Y < X must:
      (2.1) be stalled, or
      (2.2) be cleared, or
      (2.3) have WE set to 0.
    (3) (opt.) clear must be NOP operation (couldn't we assume that?)
    '''

    supress_err = True

    # Check part described in (B.1)
    # Proof of s_1 <=> ... <=> s_k using series of implications of the form:
    #   s_1 => s_2 => ... => s_k => s_1
    info("Checking consistency part (B.1) ...")
    for phase in aPhaseDict:
        pipe = [ i for i in aPhaseDict[phase] if isPipe(aFile, i[0], i[1]) ]
        for i in range(len(pipe)):
            # Index of next storage (modulo by the no. of storages)
            j = (i + 1) % len(pipe)

            # Get all required expression for analysis
            cmd_core = "(union (expr-coib 0 (clear (select %s %s))) (expr-coib 0 (clear (select %s %s))))" % \
            (getLongType(pipe[i][1]), pipe[i][0], \
             getLongType(pipe[j][1]), pipe[j][0])

            # Header for file
            smtlibstr = ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n" 
            smtlibstr += ";;; SMTOOL run for phase %s (%s => %s)\n" % \
                (pipe, pipe[i][0], pipe[j][0])

            # Get definition of relevant signals
            # At first, collect all uarch regs
            mclause = ""
            for micro in aMicros:
                mclause += "(data-out (select register %s)) " % str(micro[0])
            # Then, merge them with previously extracted expression
            q = simpleQuery(aFile, "(smtlib (union (references %s) %s))" % (cmd_core, mclause))
            smtlibstr += parseSmtlibExport(q) + "\n"

            # Get expresions that holds for given signals
            q = simpleQuery(aFile, "(smtlib %s)" % cmd_core)
            smtlibstr += parseSmtlibExport(q) + "\n"

            # Extract stall and enable signal of current storage
            cl = getCtrl(aFile, pipe[i], CCL)

            # Build assertions for SMTLib
            smtlibstr += smtlibExport("(assert (= %s #b1) )\n" % cl[0])

            # Extract stall and enable signal of next storage
            cl = getCtrl(aFile, pipe[j], CCL)

            # Build assertions for SMTLib
            smtlibstr += smtlibExport("(assert (= %s #b0) )\n" % cl[0])

            # Additional assertions based on observation of state registers
            smtlibstr += aAsserts

            # Add addtional commands for SMTLib
            smtlibstr += smtlibExport("(check-sat)\n(exit)") + "\n"

            # Run SMTTOOL
            if debug_con:
                info("SMTTOOL run for (%s => %s)" % (pipe[i][0], pipe[j][0]))
            run = runsmt(smtlibstr)
            if run != SMTUNSAT:
                error("Clearing of pipeline register '%s' (%s) in phase %d does not "
                      "imply clearing of pipeline register '%s' (%s) in the same phase!" % \
                    (getName(pipe[i]), getComment(aFile, getName(pipe[i]), getType(pipe[i])), phase, \
                     getName(pipe[j]), getComment(aFile, getName(pipe[j]), getType(pipe[j]))))
                error("SMTTOOL run for (%s => %s)" % (getName(pipe[i]), getName(pipe[j])))
                error(smtrun2str(run))
                if not supress_err:
                    supress_err = questionYesNo("Supress further consistency "
                        "errors? (Y/n)", True)
            else:
                if debug_con:
                    info("SMTTOOL Result: %s" % smtrun2str(run))
                else:
                    if debug_dots:
                        sys.stdout.write(".")
                        sys.stdout.flush()

    if not debug_con and debug_dots:
        print

    # Check part described in (B.2). Because of (A.1) and (B.1), it is enough
    # to take one representative from each phase.

    # Checking of (B.2) itself.
    #info("Checking consistency part (B.2) ...")
    #for r in aReps:
    #    # Representative from phase X.
    #    rep = aReps[r]
    #    if rep == None:
    #        continue

    #    for r1 in [ i for i in aReps if i < r ]:
    #        # Representative from phase Y.
    #        rep1 = aReps[r1]
    #        if rep1 == None:
    #            continue

    #        # Get all required expression for analysis
    #        cmd_core = "(union (expr-coib 0 (clear (select %s %s))) (expr-coib 0 (stall (select %s %s))) (expr-coib 0 (enable (select %s %s))) (expr-coib 0 (clear (select %s %s))))" % \
    #        (getLongType(rep[1]), rep[0], \
    #         getLongType(rep1[1]), rep1[0], \
    #         getLongType(rep1[1]), rep1[0], \
    #         getLongType(rep1[1]), rep1[0])

    #        # Header for file
    #        smtlibstr = ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n" 
    #        smtlibstr += ";;; SMTTOOL run for phase (cl %s => st or we or cl %s)\n" % \
    #            (rep[0], rep1[0])

    #        # Get definition of relevant signals
    #        # At first, collect all uarch regs
    #        mclause = ""
    #        for micro in aMicros:
    #            mclause += "(data-out (select register %s)) " % str(micro[0])
    #        # Then, merge them with previously extracted expression
    #        q = simpleQuery(aFile, "(smtlib (union (references %s) %s))" % (cmd_core, mclause))
    #        smtlibstr += parseSmtlibExport(q) + "\n"

    #        # Get expresions that holds for given signals
    #        q = simpleQuery(aFile, "(smtlib %s)" % cmd_core)
    #        smtlibstr += parseSmtlibExport(q) + "\n"

    #        # Extract stall and enable signal of representative from phase X.
    #        cl = getCtrl(aFile, rep, CCL)

    #        # Build assertions for SMTLib
    #        smtlibstr += smtlibExport("(assert (= %s #b1) )\n" % cl[0])

    #        # Extract stall, enable and clear signal of representative from phase Y.
    #        st = getCtrl(aFile, rep1, CST)
    #        we = getCtrl(aFile, rep1, CEN)
    #        cl = getCtrl(aFile, rep1, CCL)

    #        # Build assertions for SMTLib
    #        if st != None and cl != None:
    #            smtlibstr += smtlibExport("(assert (and (= %s #b0) (and (= %s #b1) (= %s #b0))) )\n" % (st[0], we[0], cl[0]))
    #        else:
    #            smtlibstr += smtlibExport("(assert (= %s #b1) )\n" % we[0])

    #        # Additional assertions based on observation of state registers
    #        smtlibstr += aAsserts

    #        # Add addtional commands for SMTLib
    #        smtlibstr += smtlibExport("(check-sat)\n(exit)") + "\n"

    #        # Run SMTTOOL
    #        if debug_con:
    #            info("SMTTOOL run for (%s => %s)" % (rep[0], rep1[0]))
    #        run = runsmt(smtlibstr)
    #        if run != SMTUNSAT:
    #            error("Clearing of pipeline register '%s' (%s) in phase %d does not "
    #                  "imply stalling or clearing of pipeline register '%s' (%s) in phase %d!" % \
    #                (getName(rep),  getComment(aFile, getName(rep), getType(rep)),   r, \
    #                 getName(rep1), getComment(aFile, getName(rep1), getType(rep1)), r1))
    #            error("SMTTOOL run for (%s => %s)" % (getName(rep), getName(rep1)))
    #            error(smtrun2str(run))
    #            if not supress_err:
    #                supress_err = questionYesNo("Supress further consistency "
    #                    "errors? (Y/n)", True)
    #        else:
    #            if debug_con:
    #                info("SMTTOOL Result: %s" % smtrun2str(run))
    #            else:
    #                if debug_dots:
    #                    sys.stdout.write(".")
    #                    sys.stdout.flush()

    #if not debug_con and debug_dots:
    #    print

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Utilities
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Entry point
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def consistency(aFile, aAsserts):
    info("Pre-computing often accessed data ...")
    info("    * access times of storages")
    (writes, reads, phases) = cacheAccessPhases(aFile, True)

    # Storages of each phase
    info("    * dictionary of phases")
    phaseDict = getPhaseDict(phases)

    # Compute representatives for each phase
    info("    * representatives")
    reps = getReps(aFile, phaseDict)

    # Determine micro-architectural registers
    info("    * micro-architectural registers")
    micros = getMicroArch(aFile)

    checkStall(aFile, phaseDict, reps, micros, aAsserts)
    checkClear(aFile, phaseDict, reps, micros, aAsserts)

def consistencyWrapper(aFile, aAssertFile, aTimeStat):
    try:
        hFile = open(aFile)
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aFile)
        sys.exit(1)

    try:
        hFile = open(aAssertFile)
        asserts = hFile.read()
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aAssertFile)
        sys.exit(1)

    try:
        hFile = open(aTimeStat, "w")
        hFile.write("Tool;Time\n")
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aTimeStat)

    consistency(aFile, asserts)

    time_stats = getTimeStatistics()
    info("Time: " + str(time_stats))
    try:
        hFile = open(aTimeStat, "a")
        hFile.write("rtlmanip;%.2f\n" % time_stats[0])
        hFile.write("smt;%.2f\n"      % time_stats[1])
        hFile.write("armc;%.2f\n"     % time_stats[2])
        hFile.close()
    except IOError:
        error("Cannot write time statitistics to %s!" % aTimeStat)

if __name__ == "__main__":
    if len(sys.argv) != 4:
        print "Usage: consistency.py anotated_vam micros_assert time_stats_file"
        print "     anotated_vam    Filename with anotated VAM model using DFA"
        print "     micros_assert   Additional asserts about microarch."
        print "                     registers."
        print "     time_stats_file File with time statistics."
        sys.exit(1)

    consistencyWrapper(sys.argv[1], sys.argv[2], sys.argv[3])
    finish()

