#!/usr/bin/python
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Name:         hazard.py
# Description:  Performs deep path analysis for storages with possible hazards.
# Author:       Lukas Charvat 20.3.2014
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import sys
import os
import datetime
import re

from common import *

MAX_PHASE = 100
RET_STATE = "r"
TAU_START_STATE = "q0"
SYM_ADDRESS_NAME = "sym_addr__"
ARMC_EXPORT_FILE = "rtl_artmc_exprm.ml"

stop_error = False

debug_raw = False
debug_stop = False
debug_dots = True

# Path finding is expensive therefore the computed paths are cached
path_cache = {}

def recountConditionPhases(aPhase, aPathConditions):
    res = []
    for pc in aPathConditions:
        res += [ (pc[0], pc[1] + aPhase) ]

    return res

def extractSigPaths(aFile, aSrc, aDst):
    # Try to find paths in cache first
    paths = path_cache.get((aFile, aSrc, aDst, None), None)
    if paths is not None:
        return paths

    q = simpleQuery(aFile, "(data-path (select signal %s) (select signal %s))" %
        (getName(aSrc), getName(aDst)))
    paths = parseSetArr(q)

    # Cache found paths
    path_cache[(aFile, aSrc, aDst, None)] = paths

    return paths

def extractPaths(aFile, aSrc, aDst, aConn):
    assert(aConn is not None)

    # Try to find paths in cache first
    paths = path_cache.get((aFile, aSrc, aDst, aConn), None)
    if paths != None:
        return paths

    q = simpleQuery(aFile, "(data-path (data-out (select %s %s)) (%s (select %s %s)))" % \
        (getLongType(getType(aSrc)), getName(aSrc), aConn, \
         getLongType(getType(aDst)), getName(aDst)))
    paths = parseSetArr(q)

    # Cache found paths
    path_cache[(aFile, aSrc, aDst, aConn)] = paths

    return paths

def extractPathObjects(aPath):
    path = parseTupleArr(aPath)
    objs = []
    for item in parseTupleArr(path[0]):
        if item[:4] == "(reg":
            name = ""
            i = 5
            while item[i] != " ":
                name += item[i]
                i += 1
            objs += [ (name, TREG) ]
        elif item [:3] == "(:=":
            name = ""
            i = 4
            while item[i] != " ":
                name += item[i]
                i += 1
            objs += [ (name, TSIG) ]
    return objs

def printPath(aFile, aPath):
    path = parseTupleArr(aPath)
    regs = []
    for item in parseTupleArr(path[0]):
        if item[:4] == "(reg":
            name = ""
            i = 5
            while item[i] != " ":
                name += item[i]
                i += 1
            regs += [ name ]
        #elif item [:3] == "(:=":
        #    name = "["
        #    i = 4
        #    while item[i] != " ":
        #        name += item[i]
        #        i += 1
        #    name += "]"
        #    regs += [ name ]

    info("Registers along path:")
    info("    {0:16} {1:48}".format("Register", "Comment"))
    info("    " + "-" * 71)
    for reg in regs:
        comm = getComment(aFile, reg, TREG)
        if len(comm) > 48:
            comm = "... " + comm[-44:]
        info("    {0:16} {1:48}".format(reg, comm))
    info("Path:")
    out = "   "
    for reg in regs:
        out += " -> " + reg
    out += " ->"
    info(out)

def getAddressSize(aFile, aPort):
    addr = getAddress(aFile, aPort)
    query = simpleQuery(aFile, "(bitwidth (select signal %s))" % getName(addr))
    sz = parseInt(query)
    return sz

def extractPathConditions(aFile, aPath):
    path = parseTupleArr(aPath)
    conds = []
    for t in parseSetArr(path[1]):
        val = parseTupleArr(t)
        sig_val = parseTupleArr(val[0])
        cond = ConstCond.fromstring(obj(sig_val[0], TSIG), sig_val[1])
        conds += [ (cond, int(val[1])) ]
    return conds

#def extractNopConditions(aFile, aPhaseDict, aPhase, aSkipPhases):
#    conds = []
#    for (phase, storages) in aPhaseDict.items():
#        if phase not in aSkipPhases:
#            for obj in [ i for i in storages if isPipe(aFile, getName(i), getType(i)) ]:
#                dout = getDataout(aFile, obj)
#                bw = getBitwidth(aFile, obj)
#                conds += [ (getName(dout), bw, "(_ 0 %d)" % bw) ]
#
#    return conds

#def getWriteCondition(aFile, aObj):
#    return getCtrl(aFile, aObj, CEN)

def getInfluencingSignals(aFile, aSignal, aPhase):
    q = simpleQuery(aFile, "(set s | (in s (references (expr-coib 0 (select %s %s)))) (= (getattr s \"wphs\") \"%d\"))" % (getLongType(getType(aSignal)), getName(aSignal), aPhase))

    return parseSet(q)
    
def getPipelineRegisters(aFile, aPhaseDict, aInclPhases=None, aExclPhases=None):
    return getStorages(aFile, aPhaseDict, aInclPhases, aExclPhases, PIPE_FILTER)

#    # Get all or selected phases
#    if aInclPhases != None:
#        phases = set(aInclPhases)
#    else:
#        phases = set(aPhaseDict.keys())
#
#    # Remove excluded phases
#    if aExclPhases != None:
#        phases -= set(aExclPhases)
#
#    # List of pipeline registers
#    regs = []
#
#    # For each accounted phase
#    for phase in phases:
#        regs += [ i for i in aPhaseDict[phase] if isPipe(aFile, getName(i), getType(i)) ]
#
#    return regs

def getExprDown(aFile, aSignal):
    q = simpleQuery(aFile, "(references (expr-coib 0 (select %s %s)))" % (getLongType(getType(aSignal)), getName(aSignal)))

    return parseSet(q)

def getMaximalElements(aSet, aOrd):
    max_elems = []
    for e1 in aSet:
        found = False
        for e2 in aSet:
            if e1 != e2 and aOrd(e1, e2):
                found = True
                break
        if not found:
            max_elems += [ e1 ]

    return max_elems

def getInfluencingNopConditions(aFile, aSignal, aRegs):
    reg_sigs = ""
    for reg in aRegs:
        reg_sigs += "(data-out (select register %s)) " % getName(reg)

    cmd = "(intersection (references (expr-coib 0 (select %s %s))) (union %s))" % (getLongType(getType(aSignal)), getName(aSignal), reg_sigs)
    q = simpleQuery(aFile, cmd)
    qlist = parseSet(q)

    conds = []
    for obj in qlist:
        bw = getBitwidth(aFile, obj)
        conds += [ ConstCond(obj, 0, bw)  ]
    return conds

def getSufficientPhaseConditions(aFile, aUAsserts, aUarchs, aSigCond, aPhaseSigs, aOtherConds):
    '''
    Computes sufficient conditions over a set of signals from a given phase
    implying validity of condition over the other signal (possibly computed
    using data from multiple phases).

    @param aSigCond     Necessary signal condition (computed from
                        multiple phases).
    @param aPhaseSigs   Signals of the same phase for which the sufficient
                        conditions are computed.
    @param aOtherCons   Additional conditions that should be assumed.
    @return             Set of sets of sufficient conditions over aPhaseSigs.
                        The set is empty aOtherCons are sufficient by themselves.
    '''

    # At first, collect all uarch regs
    uarch_sigs = ""
    for micro in aUarchs:
        uarch_sigs += "(data-out (select register %s)) " % str(micro[0])
 
    # Then, add stage signals
    phase_sigs = ""
    for sig in aPhaseSigs:
        phase_sigs += "(select signal %s) " % getName(sig)

    # Next, join signals from NOP conditions
    nop_sigs = ""
    for cond in aOtherConds:
        nop_sigs += "(select signal %s) " % cond.name()

    # Finally, add reference to the examined signal condition
    coib_sigs = "(references (expr-coib 0 (select signal %s)))" % aSigCond.name()

    # Query RTLManip in order to compute definition of the above
    # collected signals
    q = simpleQuery(aFile, "(smtlib (union %s %s %s %s))" % (coib_sigs, uarch_sigs, phase_sigs, nop_sigs))
    smt = parseSmtlibExport(q) + "\n"

    # Get backward COI for the examined signal
    q = simpleQuery(aFile, "(smtlib (expr-coib 0 (select signal %s)))" % aSigCond.name())
    smt += parseSmtlibExport(q) + "\n"

    # Pre-check whether the other conditions are not sufficient by themselves.
    # This can be done by checking if a negation of aSigCond is 
    # satisfiable.

    # Add negated assertion about examined signal
    pre_smt = smt
    pre_smt += smtlibExport(aSigCond.smt(aNegate=True) + " ; signal condition (negated)\n")

    # Add assertion about examined signal
    smt += smtlibExport(aSigCond.smt() + " ; signal condition\n")

    # Add assertions given by other conditions, e.g., NOP conditions, ...
    for cond in aOtherConds:
        smt += smtlibExport(cond.smt() + " ; other condition\n")
        pre_smt += smtlibExport(cond.smt() + " ; other condition\n")

    # Add assertions about uarch. registers
    smt += aUAsserts
    pre_smt += aUAsserts

    # Add additional commands for SMTLib
    pre_smt += "; Pre-printed commands for possible debugging\n"
    pre_smt += "; (apply ctx-solver-simplify) ; simplifies whole instance (Z3 only)\n"
    # Solver settings
    pre_smt += "(check-sat)\n(get-model)\n(exit)\n"

    # Get model if formula is SAT
    model = smtGetBvModel(pre_smt)
    if model == None:
        return set([])

    # Ok, the aOtherConds are not sufficient, we may proceed to find 
    # the sufficient part within aPhaseConds set.

    # Find all sat solutions
    found_models = []
    while True:
        if len(found_models) > 0:
            # We have found some model, therefore construction of
            # "not assertion" for all these models is required
            not_assert = "(assert (not\n"

            for i in range(len(found_models)):
                # Filter only relevant signals (those in aPhaseSigs set)
                na_state = []
                for sig in aPhaseSigs:
                    na_state += [ (getName(sig), found_models[i][getName(sig)]) ]

                # For each found model, we construct a formula of the form:
                # (and (= var_1 val_1) ... (and (= var_n val_n)) ... )
                ands = ""
                for j in range(len(na_state)):
                    if j == len(na_state) - 1:
                        ands += "             (= %s %s)\n" % na_state[j]
                        ands += ")" * (len(na_state) - 1)  + "\n"
                    else:
                        ands += "        (and (= %s %s)\n" % na_state[j]

                # Append "or" part, so at the end, we construct a formula
                # of the form:
                # (or andf_1 ... (or andf_n) ... ) where andf_i are
                # the above created "and" formulas.
                if i == len(found_models) - 1:
                    not_assert += "%s" % ands
                    not_assert += ")" * (len(found_models) - 1) + "\n"
                else:
                    not_assert += "    (or \n%s\n" % ands
     
            not_assert += "))\n"
        else:
            not_assert = ""

        if debug_raw:
            info("Stringified version of asserts:\n%s" % str(not_assert))

        # Common part of SMT query
        smt_check  = smt
        # Not assertion
        smt_check += "; Not assert\n" + not_assert + "\n"
        # Add additional commands for SMTLib
        smt_check += "; Pre-printed commands for possible debugging\n"
        smt_check += "; (apply ctx-solver-simplify) ; simplifies whole instance (Z3 only)\n"
        # Solver settings
        smt_check += "(check-sat)\n(get-model)\n(exit)\n"

        # Get model if formula is SAT
        model = smtGetBvModel(smt_check)
        if model == None:
            break

        # Create dictionary from obtained model
        model = dict(model)
        # FIXME: double check this ...
        #print model
        #pause()
        found_models.append(model)

    # Extract unique assertions from each model
    res = set([])
    for model in found_models:
        # Get assertion valid for this model
        conds = []
        for assign in model.items():
            if assign[0] in [ getName(sig) for sig in aPhaseSigs ]:
                # Signal is a member of aPhaseSigs
                cond = ConstCond.fromstring((assign[0], TSIG), assign[1])
                conds += [ cond ]
        res.add(frozenset(conds))

    return res

def extractReadConditions(aFile, aObj, aAddrMark, aCondMark, aPhase):
    '''
    Extracts edge conditions that necessary in order the given storage to be
    read.
    '''

    # Further computation is relevant only for memory read ports
    if getType(aObj) != TMEM or getPortDir(aFile, getName(aObj)) != TMEMRO:
        return [ [] ]

    # Read conditions (set of sets)
    rd_conds = []

    # Read time
    # TODO: Make generic ...
    rt = 0

    re_sig = getCtrl(aFile, aObj, CEN)
    ra_sig = getAddress(aFile, aObj)
    ra_bw  = getBitwidth(aFile, ra_sig)

    # TODO: Derivation of sufficient conditions for reading ...
    basic_conds = [
        (ConstCond(re_sig, 1, 1), aPhase+rt), \
        (SymCond(ra_sig, SYM_ADDRESS_NAME + aAddrMark, ra_bw), aPhase+rt) ]

    # Show statistics
    info("Got 1 sufficient condition for read from %s in %d ..." % \
        (getName(aObj), aPhase))

    # Show extracted conditions
    info("Solution 1 of 1:")
    printExecutionConditions(aFile, basic_conds)

    # Add an unique marker
    basic_conds = markPathConditions(basic_conds, aCondMark)
 
    # Adjoin the condition to the resulting set
    rd_conds += [ basic_conds ]

    return rd_conds

def extractWriteConditions(aFile, aPhaseDict, aUAsserts, aUarchs, aWrites, \
    aObj, aConn, aAddrMark, aCondMark, aPhase):
    '''
    Extracts edge conditions that necessary in order the given storage to be
    written.
    '''

    # Write conditions (set of sets)
    wr_conds = []

    # Determine write enable signal
    we_sig = getCtrl(aFile, aObj, CEN)

    # Write time
    # TODO: Make generic ...
    wt = 1

    basic_conds = []

    # Determine basic conditions
    if getType(aObj) == TMEM and getPortDir(aFile, getName(aObj)) == TMEMWO:
        if aConn == TD:
            # Create constraint over symbolic address for memory ports
            wa_sig  = getAddress(aFile, aObj)
            wa_bw   = getBitwidth(aFile, wa_sig) 
            we_cond = ConstCond.fromstring(we_sig, "(_ 1 1)")
            wa_cond = SymCond.fromstring(wa_sig, \
                SYM_ADDRESS_NAME + aAddrMark, wa_bw)
            basic_conds += [ (we_cond, aPhase-wt), (wa_cond, aPhase-wt) ]
        elif aConn == TADDR:
            we_cond = ConstCond.fromstring(we_sig, "(_ 1 1)")
            basic_conds += [ (we_cond, aPhase-wt) ]
    elif getType(aObj) == TREG:
        if aConn == TD:
            we_cond = ConstCond.fromstring(we_sig, "(_ 1 1)")
            basic_conds += [ (we_cond, aPhase-wt) ]

    # Test whether enable signal is set just by single phase
    if len(getWrites(aWrites, we_sig)) > 1:
        # If more phase are responsible for setting the enable signal, then
        # derive sufficient write conditions only over elements of the given
        # phase.

        # Set of signals of write-back phase that are responsible for
        # computation of the enable signal.
        infl_sigs = getInfluencingSignals(aFile, we_sig, aPhase - wt)

        # Computation of the COI-Down mapping st. COI-Down(s) = Expr-Coib(s)
        # for the previously calculated set of influencing signals.
        coi_downs = [ (sig, set(getExprDown(aFile, sig))) for sig in infl_sigs ]
        # Find maximal elements among item of the downward mapping.
        max_elems = getMaximalElements(coi_downs, \
            lambda (s1, d1), (s2, d2): d1.issubset(d2))

        # For our purpose, it is enough to use only the computed
        # maximal elements.
        infl_sigs = [ i[0] for i in max_elems ]

        # Signals from other phases are supposed to be in the same
        # state as they would be during execution of NOP instruction.
        pipe_regs = getPipelineRegisters(aFile, aPhaseDict, None, [ aPhase-wt ])
        infl_nop_conds = getInfluencingNopConditions(aFile, we_sig, pipe_regs)

        # Now, compute sufficient conditions themselves.
        suff_conds = getSufficientPhaseConditions(aFile, aUAsserts, aUarchs, \
            ConstCond(we_sig, 1, getBitwidth(aFile, we_sig)), infl_sigs, \
            infl_nop_conds)

        # Show statistics
        sol_tot = len(suff_conds)
        if sol_tot == 0:
            # Use at least basic conditions if there are any
            if basic_conds == []:
                return [ [] ]

            # Show statistics
            info("Got 1 sufficient condition for write to %s in %d ..." % \
                (getName(aObj), aPhase))

            # Show extracted conditions
            info("Solution 1 of 1:")
            printExecutionConditions(aFile, basic_conds)

            # Add an unique marker
            basic_conds = markPathConditions(basic_conds, aCondMark)

            # Adjoin the condition to the resulting set
            wr_conds += [ basic_conds ]
        else:
            sol_cnt = 0
            info("Got %d sufficient condition(s) for write to %s in %d ..." % \
                (sol_tot, getName(aObj), aPhase))
 
            # Each set of sufficient conditions 
            for conds in suff_conds:
                sol_cnt += 1
                new_conds = basic_conds + \
                    [ (cond, aPhase-wt) for cond in conds ]

                # Make elements unique
                new_conds = list(set(new_conds))
 
                # Show extracted conditions
                info("Solution %d of %d:" % (sol_cnt, sol_tot))
                printExecutionConditions(aFile, new_conds)

                # Add an unique marker
                new_conds = markPathConditions(new_conds, aCondMark)
 
                # Adjoin the condition to the resulting set
                wr_conds += [ new_conds ]
    else:
        # No need to compute sufficient conditions here.

        if basic_conds == []:
            return [ [] ]

        # Show statistics
        info("Got 1 sufficient condition for write to %s in %d ..." % \
            (getName(aObj), aPhase))

        # Show extracted conditions
        info("Solution 1 of 1:")
        printExecutionConditions(aFile, basic_conds)

        # Add an unique marker
        basic_conds = markPathConditions(basic_conds, aCondMark)

        # Adjoin the condition to the resulting set
        wr_conds += [ basic_conds ]

    return wr_conds

def getForwardHazardCases(aFile, aPaths, aReadTime, aWriteTime, aWrites, aReads):
    # if aWriteTime - aReadTime <= 1:
    #    return []
    # FIXME: Support for timing other than (0,1) ...
    fhcs = []
    for path in aPaths:
        path_objs = extractPathObjects(path)
        #info("Path: %s" % str(path_objs))

        n = aReadTime
        path_valid = True
        for obj in path_objs:
            if getType(obj) == TREG:
                n += 1 

            #info("N: %d" % n)
            #info("Obj: %s t_w: %s" % (str(obj), \
            #    str(aWrites[obj])))
            #info("Obj: %s t_r: %s" % (str(obj), \
            #    str(aWrites[obj])))

            # No write stage means no restriction for signal
            if aWrites[obj] != [] and not n in aWrites[obj]:
                path_valid = False
                #info("FWD: Eliminating path:")
                #info("Path: %s" % str(path_objs))
                break

        if path_valid and n+1 == aWriteTime:
            path_conds = extractPathConditions(aFile, path)
            path_conds = recountConditionPhases(aReadTime, path_conds)
            fhcs += [ (path_conds, path_objs) ]
            if debug_dots:
                sys.stdout.write("+")
                sys.stdout.flush()
        #else:
        #    if debug_dots:
        #        sys.stdout.write("-")
        #        sys.stdout.flush()
            
    if debug_dots:
        print

    return fhcs

def getTurnHazardCases(aFile, aPaths, aReadTime, aWriteTime, aWrites, aReads):
    thcs = []
    p = 0
    path_cache = []
    for path in aPaths:
        p += 1
        path_objs = extractPathObjects(path)
        path_conds = extractPathConditions(aFile, path)
        path_conds = recountConditionPhases(aReadTime, path_conds)

        n = aReadTime
        prev_reg = None
        turn_objs = []
        prev_turn_objs = []

        found = False
        valid = True
        for i in range(len(path_objs)):
            if turn_objs in path_cache:
                continue

            obj = path_objs[i]
            turn_objs += [ obj ]

            if getType(obj) == TREG:
                n += 1
                prev_reg = obj
                prev_turn_objs = turn_objs + []

            #info("N: %d" % n)
            #info("%s: %s t_w: %s" % (getType(obj), getName(obj), \
            #    str(aWrites[obj])))
            #info("%s: %s t_r: %s" % (getType(obj), getName(obj), \
            #    str(aReads[obj])))
            #info("%s %s" % (str(turn_objs), str(path_cache)))
            #pause()

            if not found:
                if not n in aWrites[obj] or not n in aReads[obj]:
                    turn_conds = [ c for c in path_conds if c[1] < n - 1 ]
                    thc = (prev_reg, n, turn_conds, prev_turn_objs)
                    found = True
                    # TODO: Mark that we have found turning point, however 
                    # it can be still invalid. So analysis has to continue.
                    reads = aReads[obj]
                    if reads == []:
                        valid = False
                        break
                    else:
                        n = min(reads)
            else:
                if not n in aReads[obj]:
                    reads = aReads[obj]
                    if reads == []:
                        valid = False
                        break
                    else:
                        n = min(reads)

        # Turning point must be found and path has to be valid
        if found and valid and n + 1 == aWriteTime:
            #info("Turning point found at %s in time %d" % (thc[0], thc[1]))
            #info("Retained path: %s" % str(thc[3]))
            #info("Retained path conditions: %s" % str(thc[2]))
            #pause()
            if not thc in thcs:
                thcs += [ thc ]
            if not thc[3] in path_cache:
                path_cache += [ thc[3] ]
            if debug_dots:
                sys.stdout.write(".")
                sys.stdout.flush()

    if debug_dots:
        print
 
    return thcs

def printRAWHC(aFile, aHC):
    info("RAW Hazard case description:")
    info("    To:   %s %s" % (getName(aHC[0]), getComment(aFile, \
        getName(aHC[0]), getType(aHC[0]))))
    info("    At:   %d" % aHC[1])
    info("    From: %s %s" % (getName(aHC[2]), getComment(aFile, \
        getName(aHC[2]), getType(aHC[2]))))
    info("    At:   %d" % aHC[3])
    info("    To:   %s (%s) %s" % (getName(aHC[4][0]), aHC[4][1], 
        getComment(aFile, getName(aHC[4][0]), getType(aHC[4][0]))))
    info("    At:   %d" % aHC[5])
    info("    H:    %d" % (aHC[5] - aHC[3]))
    info("    Path Objects:")
    info("    {0:16} {1:8} {2:48}".format("Name", "Type", "Comment"))
    info("    " + "-" * 71)
    for obj in aHC[7]:
        comm = getComment(aFile, getName(obj), getType(obj))
        if len(comm) > 32:
            comm = "... " + comm[-28:]
        info("    {0:16} {1:8} {2:48}".format(getName(obj), getType(obj), comm))

def printWAWHC(aFile, aHC):
    info("WAW Hazard case description:")
    info("    To:   %s %s" % (getName(aHC[0]), getComment(aFile, \
        getName(aHC[0]), getType(aHC[0]))))
    info("    At:   %d" % aHC[1])
    info("    To:   %s %s" % (getName(aHC[2]), getComment(aFile, \
        getName(aHC[2]), getType(aHC[2]))))
    info("    At:   %d" % aHC[3])
    info("    H:    %d" % (aHC[3] - aHC[1]))

def printWARHC(aFile, aHC):
    info("WAR Hazard case description:")
    info("    To:   %s %s" % (getName(aHC[4]), getComment(aFile, \
        getName(aHC[4]), getType(aHC[4]))))
    info("    At:   %d" % aHC[5])
    info("    From: %s %s" % (getName(aHC[0]), getComment(aFile, \
        getName(aHC[0]), getType(aHC[0]))))
    info("    At:   %d" % aHC[1])
    info("    To:   %s (%s) %s" % (getName(aHC[2][0]), aHC[2][1], 
        getComment(aFile, getName(aHC[2][0]), getType(aHC[2][0]))))
    info("    At:   %d" % aHC[3])

    info("    H:    %d" % (aHC[1] - aHC[5]))
    info("    Path Objects:")
    info("    {0:16} {1:8} {2:48}".format("Name", "Type", "Comment"))
    info("    " + "-" * 71)
    for obj in aHC[7]:
        comm = getComment(aFile, getName(obj), getType(obj))
        if len(comm) > 32:
            comm = "... " + comm[-28:]
        info("    {0:16} {1:8} {2:48}".format(getName(obj), getType(obj), comm))

def printCTLHC(aFile, aHC):
    info("CTL Hazard case description:")
    info("    To:   %s @%d %s" % (getName(aHC[0]), aHC[1], getComment(aFile,
        getName(aHC[0]), getType(aHC[0]))))
    info("    To:   %s @%d %s" % (getName(aHC[2][0]), aHC[3], getComment(aFile,
        getName(aHC[2][0]), getType(aHC[2][0]))))
    dist = aHC[1] - aHC[3]
    col = aHC[1] - 1
    h2 = max(dist, 1)
    h1 = col - max(dist, 1) + 1
    info("    Dist: %d    Hbr: %d    Hwr: %d" % (dist, h1, h2))
    info("    Path Objects:")
    info("    {0:16} {1:8} {2:48}".format("Name", "Type", "Comment"))
    info("    " + "-" * 71)
    for obj in aHC[5]:
        comm = getComment(aFile, getName(obj), getType(obj))
        if len(comm) > 32:
            comm = "... " + comm[-28:]
        info("    {0:16} {1:8} {2:48}".format(getName(obj), getType(obj), comm))


def markPathConditions(aPathConditions, aMark):
    res = []
    for pc in aPathConditions:
        res += [ (aMark, pc[0], pc[1]) ]

    return res

#def rollbackExecution(aFile, aMark, Q, C, R, U, aFrom, aK):
#    for i in range(1, aK+1):
#        # Get current state
#        q = Q[aFrom-i]
#        sigs = [ s[0] for s in C[q] ] + [ s[0] for s in R[q] ]
#        #      ^^^^^^^^^^^^^^^^^^^^^^   ^^^^^^^^^^^^^^^^^^^^^^
#        #           state conds         additional reg D sigs
#
#        if len(sigs) == 0:
#            continue
#
#        # Extract declarations
#        cmd = "(smtlib-prime %d (union "
#        for v in sigs:
#            cmd += "(references (expr-coib 0 (select signal %s))) (select signal %s) " % \
#                (v, v)
#        cmd += ")) "
#
#        # Extract equalities
#        cmd += "(smtlib-prime %d (union "
#        for v in sigs:
#            cmd += "(expr-coib 0 (select signal %s)) " % v
#        cmd += ")) "
#
#        # Extract registers
#        cmd += "(smtlib-prime %d (set r where (in r registers) (intersection \
#            (data-out r) (union "
#        for v in sigs:
#            cmd += "(references (expr-coib 0 (select signal %s))) (select signal %s) " % \
#                (v, v)
#        cmd += ")))) "
#
#        # Extract D signals of registers
#        cmd1 = "(set r where (in r registers) (intersection \
#            (data-out r) (union "
#        for v in sigs:
#            cmd1 += "(references (expr-coib 0 (select signal %s))) (select signal %s) " % \
#                (v, v)
#        cmd1 += ")))"
#        #print cmd1
#        query = simpleQuery(aFile, cmd1)
#        #print query
#        qlist = parseSet(query)
#        cmd1 = "(union "
#        for reg in qlist:
#            cmd1 += "(data-in (select register %s)) " % getName(reg)
#        cmd1 += ")"
#        query = simpleQuery(aFile, cmd1)
#        regs = parseSet(query)
#
#        # Append addtional "sensitive" signals to previous state
#        R[Q[aFrom-i-1]] += regs
#        #Q[aFrom-i-1] = (Q[aFrom-i-1][0], Q[aFrom-i-1][1], Q[aFrom-i-1][2], \
#        #     Q[aFrom-i-1][3] + regs, Q[aFrom-i-1][4])
#        #print regs
#
#        # For previous states
#        for j in range(i, aK+1):
#            # Append primed equations to previous states
#            smt = ";;; Begin of unroll %d (%s)\n" % (j, aMark)
#            shift = j - i + 1
#            query = simpleQuery(aFile, cmd % (shift, shift, shift - 1))
#            smt += parseSmtlibExport(query) + "\n"
#            for cond in C[q]:
#                smt += "(assert (= %s" % cond[0]
#                smt += "_" + "p" * shift
#                smt += " %s))\n" % re.sub("_ ", "_ bv", cond[1])
#            #print "i:", i, "j:", j, "-j", -j, "shift:", shift
#            smt += ";;; End of unroll %d (%s)\n" % (j, aMark)
#
#            # Add marker to primed variables
#            smt = re.sub(r"(_pp*)", r"\1_" + aMark, smt)
#            U[Q[aFrom-j-1]] = smt + U[Q[aFrom-j-1]]
#            #Q[aFrom-j-1] = (Q[aFrom-j-1][0], Q[aFrom-j-1][1], Q[aFrom-j-1][2], \
#            #    Q[aFrom-j-1][3], smt + Q[aFrom-j-1][4])

# def rollbackExecution(aFile, aMark, Q, C, R, U, aFrom, aK):
#     for i in range(1, aK+1):
#         # Get current state
#         q = Q[aFrom-i]
#         sigs = [ s[0] for s in C[q] ] + [ s[0] for s in R[q] ]
#         #      ^^^^^^^^^^^^^^^^^^^^^^   ^^^^^^^^^^^^^^^^^^^^^^
#         #           state conds         additional reg D sigs
#
#         if len(sigs) == 0:
#             continue
#
#         # Extract declarations
#         cmd = "(smtlib-prime %d (union "
#         for v in sigs:
#             cmd += "(references (expr-coib 0 (select signal %s))) (select signal %s) " % \
#                 (v, v)
#         cmd += ")) "
#
#         # Extract equalities
#         cmd += "(smtlib-prime %d (union "
#         for v in sigs:
#             cmd += "(expr-coib 0 (select signal %s)) " % v
#         cmd += ")) "
#
#         # Extract registers
#         cmd += "(smtlib-prime %d (set r where (in r registers) (intersection \
#             (data-out r) (union "
#         for v in sigs:
#             cmd += "(references (expr-coib 0 (select signal %s))) (select signal %s) " % \
#                 (v, v)
#         cmd += ")))) "
#
#         # Extract D signals of registers
#         cmd1 = "(set r where (in r registers) (intersection \
#             (data-out r) (union "
#         for v in sigs:
#             cmd1 += "(references (expr-coib 0 (select signal %s))) (select signal %s) " % \
#                 (v, v)
#         cmd1 += ")))"
#         #print cmd1
#         query = simpleQuery(aFile, cmd1)
#         #print query
#         qlist = parseSet(query)
#         cmd1 = "(union "
#         for reg in qlist:
#             cmd1 += "(data-in (select register %s)) " % getName(reg)
#         cmd1 += ")"
#         query = simpleQuery(aFile, cmd1)
#         regs = parseSet(query)
#
#         # Append addtional "sensitive" signals to previous state
#         R[Q[aFrom-i-1]] += regs
#         #Q[aFrom-i-1] = (Q[aFrom-i-1][0], Q[aFrom-i-1][1], Q[aFrom-i-1][2], \
#         #     Q[aFrom-i-1][3] + regs, Q[aFrom-i-1][4])
#         #print regs
#
#         # For previous states
#         for j in range(i, aK+1):
#             # Append primed equations to previous states
#             smt = ";;; Begin of unroll %d (%s)\n" % (j, aMark)
#             shift = j - i + 1
#             query = simpleQuery(aFile, cmd % (shift, shift, shift - 1))
#             smt += parseSmtlibExport(query) + "\n"
#             for cond in C[q]:
#                 smt += "(assert (= %s" % cond[0]
#                 smt += "_" + "p" * shift
#                 smt += " %s))\n" % re.sub("_ ", "_ bv", cond[1])
#             #print "i:", i, "j:", j, "-j", -j, "shift:", shift
#             smt += ";;; End of unroll %d (%s)\n" % (j, aMark)
#
#             # Add marker to primed variables
#             smt = re.sub(r"(_pp*)", r"\1_" + aMark, smt)
#             U[Q[aFrom-j-1]] = smt + U[Q[aFrom-j-1]]
#             #Q[aFrom-j-1] = (Q[aFrom-j-1][0], Q[aFrom-j-1][1], Q[aFrom-j-1][2], \
#             #    Q[aFrom-j-1][3], smt + Q[aFrom-j-1][4])

def buildPlainTS(aFile, aPhaseDict, aMark, aMinPhase, aMaxPhase, aDefaultFlags=""):
    # State set
    Q = [ ]
    # Execution condition mapping
    C = {}

    # Unique starting state
    qin = (aMark, aMinPhase - 1, "i" + aDefaultFlags)
    #      ^^^^^  ^^^^^^^^^^^^^  ^^^^^
    #      class  phase          flags
    Q += [ qin ]
    C[qin] = []

    # Construct additional states and collect their execution conditions
    for phase in range(aMinPhase, aMaxPhase + 1):
        # Append new state to the state set
        q = (aMark, phase, aDefaultFlags)
        Q += [ q ]
        C[q] = [ ]

    # Unique final state
    qout = (aMark, aMaxPhase + 1, "o" + aDefaultFlags)
    Q += [ qout ]
    C[qout] = []

    # Transition relation
    Delta = [ ]

    # Compute basic transitions
    qlast = None
    if len(Q) > 0:
        for i in range(len(Q)):
            # Guard set
            G = []
            # Compute all state that are in phase i + 1
            for j in range(i + 1, len(Q)):
                G += [ Q[j] ]
            # Transition from previous state
            if qlast != None:
                Delta += [ (qlast, G, Q[i]) ]
            # Make self-loop for initial state
            if i == 0:
                G1 = G + [ ]
                Delta += [ (Q[i], G1, Q[i]) ]
            # Store previous state
            qlast = Q[i]

    # Transitions of final state
    for t in Delta:
        if t[2] == qout:
            t[1].append(qout)

    # Make self-loop for final state
    Delta += [ (qout, [ qout ], qout) ]

    return (Q, Delta, qin, C, None)

def buildCondMappedTS(aFile, aPhaseDict, aPathConds, aMark, aMinPhase, aMaxPhase, aDefaultFlags=""):
    # State set
    Q = [ ]
    # Execution condition mapping
    C = {}

    # Unique starting state
    qin = (aMark, aMinPhase - 1, "i" + aDefaultFlags)
    #      ^^^^^  ^^^^^^^^^^^^^  ^^^^^
    #      class  phase          flags
    Q += [ qin ]
    C[qin] = []

    # Construct additional states and collect their execution conditions
    for phase in range(aMinPhase, aMaxPhase + 1):
        # Gather condition from the same phase
        conds = [ i[1] for i in aPathConds if i[2] == phase ]
        # Append new state to the state set
        q = (aMark, phase, aDefaultFlags)
        Q += [ q ]
        C[q] = conds

    # Unique final state
    qout = (aMark, aMaxPhase + 1, "o" + aDefaultFlags)
    Q += [ qout ]
    C[qout] = []

    # Transition relation
    Delta = [ ]

    # Compute basic transitions
    qlast = None
    if len(Q) > 0:
        for i in range(len(Q)):
            # Guard set
            G = []
            # Compute all state that are in phase i + 1
            for j in range(i + 1, len(Q)):
                G += [ Q[j] ]
            # Transition from previous state
            if qlast != None:
                Delta += [ (qlast, G, Q[i]) ]
            # Make self-loop for initial state
            if i == 0:
                G1 = G + [ ]
                Delta += [ (Q[i], G1, Q[i]) ]
            # Store previous state
            qlast = Q[i]

    # Transitions of final state
    for t in Delta:
        if t[2] == qout:
            t[1].append(qout)

    # Make self-loop for final state
    Delta += [ (qout, [ qout ], qout) ]

    #U = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, [], Q[1:-1], Q[1],
    #    C, COI_BCK, aMark)

    return (Q, Delta, qin, C, None)

def mergeTS(aTS1, aTS2, aCheckSanity=True):
    '''
    Performs merge of aTS2 into aTS1. Unrolled phase conditions
    and input state are kept for aTS1 only.
    '''

    (Q1, D1, q1in, C1, U1) = aTS1
    (Q2, D2, q2in, C2, U2) = aTS2

    # Check whether merging of these two TSs is possible
    if aCheckSanity and any(set(Q1).intersection(set(Q2))):
        error("States of transition system have non-empty intersection!")
        exit(1)

    # Perform actual merge
    Q3 = Q1 + Q2
    D3 = D1 + D2
    q3in = q1in
    C3 = C1.copy()
    C3.update(C2)
    U3 = U1

    return (Q3, D3, q3in, C3, U3)

def hasHazardFlag(aQ, aN):
    (mark, phase, flags) = aQ

    return (flags.find(str(aN)) != -1)

def unrollAlongPathTS(aFile, aPhaseDict, aTS, aPath, aMark, aSanityCheck=True):
    (Q, D, qin, C, U) = aTS

    if aSanityCheck:
        if not U is None:
            error("A given system has already unrolled conditions!")
            exit(1)
        if len(aPath) < 1:
            error("A given path is too short!")
            exit(1)
        for q in aPath:
            if not q in Q:
                error("A given path is not present in a given system!")
                exit(1)

    U = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, [], aPath, aPath[0],
        C, COI_BCK, aMark)

    return (Q, D, qin, C, U)

def flagHazardousTS(aTS, aQh, aCount, aPrefix, aOtherTS):
    '''
    Expands states of a given transition system so it contains
    a provided number of hazardous states.
    '''

    (Q, D, qin, C, U) = aTS
    (mark_h,  phase_h,  flags_h ) = aQh
    (mark_in, phase_in, flags_in) = qin
    (Qo, Do, qino, Co, Uo) = aOtherTS

    if aQh not in Q:
        error("Entry hazard state is not in the set of states!")
        exit(1)

    if aCount < 1:
        error("The number of hazard states is less than one!")
        exit(1)

    if not U is None:
        warning("System has non-empty unrolled path conditions, they will be lost!")

    if not Uo is None:
        warning("System has non-empty unrolled path conditions, they will be lost!")

    # Set of new states
    Qnew = [ ]
    Cnew = { }
    qin_new = (mark_in, phase_in, flags_in + aPrefix + "0")

    # Rename states (i.e., make cartesian product with hazard state numbers)
    for q in Q:
        (mark, phase, flags) = q
        for i in range(aCount + 1):
            qnew = (mark, phase, flags + aPrefix + str(i))
            Qnew += [ qnew ]
            Cnew[qnew] = C[q]

    # New transition of other system
    Do_new = [ ]
    for (q1, Go, q2) in Do:
        Go_new = [ ]
        for (mark, phase, flags) in Go:
            if (mark, phase, flags) in Q:
                for j in range(aCount + 1):
                    Go_new += [ (mark, phase, flags + aPrefix + str(j)) ]
            else:
                Go_new += [ (mark, phase, flags) ]
        Do_new += [ (q1, Go_new, q2) ]

    # New transition relation
    Dnew = [ ]

    # Incrementally build the new transition relation
    # 1. Include original transitions
    for ((mark1, phase1, flags1), G, (mark2, phase2, flags2)) in D:
        # If the originating state is NOT equal to hazard entry state
        if not (mark1 == mark_h) or not (phase1 == phase_h) or not (flags1 == flags_h):
            # Update guard set
            Gnew = [ ]
            for (mark, phase, flags) in G:
                if (mark, phase, flags) in Q:
                    for j in range(aCount + 1):
                        Gnew += [ (mark, phase, flags + aPrefix + str(j)) ]
                else:
                    Gnew += [ (mark, phase, flags) ]
            Dnew += [ ((mark1, phase1, flags1 + aPrefix + "0"), Gnew, (mark2, phase2, flags2 + aPrefix + "0")) ]

    # 2. Inter-hazardous transitions
    for ((mark1, phase1, flags1), G, (mark2, phase2, flags2)) in D:
        for i in range(1, aCount):
            # Update guard set
            Gnew = [ ]
            for (mark, phase, flags) in G:
                if (mark, phase, flags) in Q:
                    for j in range(aCount + 1):
                        Gnew += [ (mark, phase, flags + aPrefix + str(j)) ]
                else:
                    Gnew += [ (mark, phase, flags) ]
            Dnew += [ ((mark1, phase1, flags1 + aPrefix + str(i)), Gnew, (mark2, phase2, flags2 + aPrefix + str(i+1))) ]

    # 3. Start of the hazard
    for ((mark1, phase1, flags1), G, (mark2, phase2, flags2)) in D:
        # If the originating state is equal to hazard entry state
        if (mark1 == mark_h) and (phase1 == phase_h) and (flags1 == flags_h):
            # Update guard set
            Gnew = [ ]
            for (mark, phase, flags) in G:
                if (mark, phase, flags) in Q:
                    for j in range(aCount + 1):
                        Gnew += [ (mark, phase, flags + aPrefix + str(j)) ]
                else:
                    Gnew += [ (mark, phase, flags) ]
            Dnew += [ ((mark1, phase1, flags1 + aPrefix + "0"), Gnew, (mark2, phase2, flags2 + aPrefix + "1")) ]

    # 4. Ends of hazards
    for ((mark1, phase1, flags1), G, (mark2, phase2, flags2)) in D:
        # Update guard set
        Gnew = [ ]
        for (mark, phase, flags) in G:
            if (mark, phase, flags) in Q:
                for j in range(aCount + 1):
                    Gnew += [ (mark, phase, flags + aPrefix + str(j)) ]
            else:
                Gnew += [ (mark, phase, flags) ]
        Dnew += [ ((mark1, phase1, flags1 + aPrefix + str(aCount)), Gnew, (mark2, phase2, flags2 + aPrefix + "0")) ]

    ts = (Qnew, Dnew, qin_new, Cnew, None)
    ts_other = (Qo, Do_new, qino, Co, None)

    return (ts, ts_other)

def eliminateUnreachableTS(aTS, aOtherTS):
    '''
    Eliminates unreachable states of a given transition system.
    '''

    (Q, D, qin, C, U) = aTS
    (Qo, Do, qino, Co, Uo) = aOtherTS

    if not U is None:
        warning("System has non-empty unrolled path conditions, they will be lost!")

    Reachable = set([])
    Reachable_new = set([ qin ])
    # While iteratively constructed set is not equal to the originating one
    while any(Reachable_new.symmetric_difference(Reachable)):
        Reachable = Reachable_new
        Reachable_new = set([ qin ])
        for q in Reachable:
            for (q1, G, q2) in D:
                if q1 == q:
                    Reachable_new.add(q2)

    # Trasition relation of other TS
    Do_new = [ ]
    for (q1, Go, q2) in Do:
        # Keep only reachable guards or those which
        # are NOT part of the currently simplified TS
        Go_new = [ q for q in Go if q in Reachable or not q in Q ]
        Do_new += [ (q1, Go_new, q2) ]

    # Prune transition relation
    Dnew = [ ]
    for (q1, G, q2) in D:
        if q1 in Reachable:
            Gnew = [ ]
            for q in G:
                # Keep only reachable guards or those which
                # are NOT part of the currently simplified TS
                if q in Reachable or not q in Q:
                    Gnew += [ q ]
            Dnew += [ (q1, Gnew, q2) ]

    return ((sorted(list(Reachable)), sorted(Dnew), qin, C, None), (Qo, Do_new, qino, Co, Uo))

def buildBasicTS(aFile, aPhaseDict, aPathConds, aMark, aMinPhase, aMaxPhase, aHazardStart, aHazardEnd, aReadPhase, aLoops):
    # State set
    Q = [ ]
    # Execution condition mapping
    C = {}

    # Unique starting state
    qin = (aMark, aMinPhase - 1, "i"  )
    #      ^^^^^  ^^^^^^^^^^^^^  ^^^^^
    #      class  phase          flags
    Q += [ qin ]

    # Construct additional states and collect their execution conditions
    for phase in range(aMinPhase, aMaxPhase + 1):
        # Gather condition from the same phase
        #print "cond", [ i[1] for i in aPathConds if i[2] == phase ]
        conds = [ i[1] for i in aPathConds if i[2] == phase ]
        # Append new state to the state set
        flags = ""
        if phase >= aHazardStart and phase < aHazardEnd:
            flags += "h" + str(phase - aHazardStart)
        if phase == aReadPhase:
            flags += "r"

        q = (aMark, phase, flags)
        Q += [ q ]
        C[q] = conds

    # Final states (hazardous)
    for i in range(aMaxPhase + 1, aHazardEnd):
        q = (aMark, i, "oh" + str(i - aHazardStart))
        Q += [ q ]
        C[q] = []

    # Unique final state
    qout = (aMark, max(aMaxPhase + 1, aHazardEnd), "o")
    Q += [ qout ]

    # Transitions relation
    Delta = [ ]

    # Compute basic transitions
    qlast = None
    if len(Q) > 0:
        for i in range(len(Q)):
            # Guard set
            G = []
            # Compute all state that are in phase i + 1
            for j in range(i + 1, len(Q)):
                G += [ Q[j] ]
            # Transition from previous state
            if qlast != None:
                Delta += [ (qlast, G, Q[i]) ]
            # Make self-loop for initial state
            if i == 0 or aLoops:
                G1 = G + [ ]
                Delta += [ (Q[i], G1, Q[i]) ]
            # Store previous state
            qlast = Q[i]

    # Transitions of final state
    for t in Delta:
        if t[2] == qout:
            t[1].append(qout)

    # Make self-loop for final state
    Delta += [ (qout, [ qout ], qout) ]

    U = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, [], Q[1:-1], Q[1],\
        C, COI_BCK, aMark)

    return (Q, Delta, qin, C, U)

#def buildBasicTS(aFile, aPathConds, aMark, aMinPhase, aMaxPhase, aHazardStart, aHazardEnd, aReadPhase, aLoops):
#    # State set
#    Q = [ ]
#    # Execution condition mapping
#    C = {}
#    # Register info mapping
#    R = {}
#    # Unroll mapping
#    U = {}
#
#    # Unique starting state
#    qin = (aMark, aMinPhase - 1, "i"  )
#    #      ^^^^^  ^^^^^^^^^^^^^  ^^^^^
#    #      class  phase          flags
#
#    Q += [ qin ]
#    C[qin] = []
#    R[qin] = []
#    U[qin] = ""
#
#    # Construct additional states and collect their execution conditions
#    for phase in range(aMinPhase, aMaxPhase + 1):
#        # Gather condition from the same phase
#        conds = [ i[1] for i in aPathConds if i[2] == phase ]
#        # Append new state to the state set
#        flags = ""
#        if phase >= aHazardStart and phase < aHazardEnd:
#            flags += "h" + str(phase - aHazardStart)
#        if phase == aReadPhase:
#            flags += "r"
#
#        q = (aMark, phase, flags)
#        Q += [ q ]
#        C[q] = conds
#        R[q] = []
#        U[q] = ""
#
#    # Find the first state with non
#    k = aMaxPhase - aMinPhase
#    rollbackExecution(aFile, aMark, Q, C, R, U, len(Q), k)
#
#    # Final states (hazardous)
#    for i in range(aMaxPhase + 1, aHazardEnd):
#        q = (aMark, i, "oh" + str(i - aHazardStart))
#        Q += [ q ]
#        C[q] = []
#        R[q] = []
#        U[q] = ""
#
#    # Unique final state
#    qout = (aMark, max(aMaxPhase + 1, aHazardEnd), "o")
#    Q += [ qout ]
#    C[qout] = []
#    R[qout] = []
#    U[qout] = ""
#
#    # Transitions relation
#    Delta = [ ]
#
#    # Compute basic transitions
#    qlast = None
#    if len(Q) > 0:
#        for i in range(len(Q)):
#            # Guard set
#            G = []
#            # Compute all state that are in phase i + 1
#            for j in range(i + 1, len(Q)):
#                G += [ Q[j] ]
#            # Transition from previous state
#            if qlast != None:
#                Delta += [ (qlast, G, Q[i]) ]
#            # Make self-loop for initial state
#            if i == 0 or aLoops:
#                G1 = G + [ ]
#                Delta += [ (Q[i], G1, Q[i]) ]
#            # Store previous state
#            qlast = Q[i]
#
#    # Transitions of final state
#    for t in Delta:
#        if t[2] == qout:
#            t[1].append(qout)
#
#    # Make self-loop for final state
#    Delta += [ (qout, [ qout ], qout) ]
#
#    return (Q, Delta, qin, C, R, U)

def completeGuards(aTS1, aTS2):
    Q1 = aTS1[0]
    Q2 = aTS2[0]

    Delta1 = aTS1[1]
    Delta2 = aTS2[1]

    for d in Delta1:
        for q in Q2:
            if d[2][1] < q[1] or q[2].find("o") != -1:
                if not q in d[1]:
                    d[1].append(q)

    for d in Delta2:
        for q in Q1:
            if d[2][1] < q[1] or q[2].find("o") != -1:
                if not q in d[1]:
                    d[1].append(q)

def printTS((Q, Delta, qin, C, U)):
    info("Transition system:")
    info("States:")
    info("    {0:8} {1:8} {2:52}".format("Symname", "Phase", "Condition"))
    info("    " + "-" * 74)
    for q in Q:
        cond = str(C.get(q, []))
        #if len(cond) > 52:
        #    cond = cond[:49] + "..."
        info("    {0:8} {1:8} {2:52}".format(q2str(q), str(q[1]), cond))
    info("Transitions:")
    info("    {0:8} {1:8} {2:52}".format("Source", "Dest.", "Guard"))
    info("    " + "-" * 74)
    for t in Delta:
        guard = ""
        for g in t[1]:
            guard += q2str(g) + " "
        info("    {0:8} {1:8} {2:52}".format(q2str(t[0]), q2str(t[2]), guard))
    info("Input state:")
    info("    " + q2str(qin))
    #info("R U:")
    #for q in Q:
    #    info("    {0:8} {1:8} {2:52}".format(q2str(q), str(R[q]), str(U[q])))

def q2str((aMark, aPhase, aFlags)):
    return aMark + str(aPhase) + aFlags

def buildTransducer(TSs):
    # States of a transducer are transitions of input transition systems
    Q = []
    for TS in TSs:
        Q += TS[1]

    Delta = []
    for q1 in Q:
        for q2 in Q:
            if q1[2] in q2[1]:
                Delta = Delta + [ (q1, q2) ]

    # Renaming of states
    Qnew = []
    for q in Q:
        q1new = q2str(q[0])
        q2new = q2str(q[2])
        Qnew = Qnew + [ (q1new, q2new) ]

    Deltanew = []
    for t in Delta:
        q11new = q2str(t[0][0])
        q12new = q2str(t[0][2])
        q21new = q2str(t[1][0])
        q22new = q2str(t[1][2])
        tnew = ((q11new, q12new), (q21new, q22new))
        Deltanew = Deltanew + [ tnew ]

    return (Qnew, Deltanew, Qnew)

def saveARMC(aTarget, aContent):
    try:
        hFile = open(aTarget, "w")
        hFile.write(aContent)
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aTarget)

def sanitize(aStr):
    res = re.sub("[-\[\]\+ ]","_",aStr)
    return res

def headerARMC(aFile):
    out =       "(* CPU verification export from ARTMC *)\n"
    out = out + "(* Original input file: %s *)\n" % str(aFile)
    out = out + "(* Date: %s *)\n\n" % str(datetime.datetime.now())

    out = out + "open Taml;;\n"
    out = out + "open Dxn;;\n"
    out = out + "open Colapsing_v3;;\n\n"

    return out

def footerARMC():
    return  "atrmc_strpres_fwcomp_bwcoll_allstpred sigma_str init_str tau_str bad_str [ sigstar_str ];()"

def exportSigmaARMC(M):
    out = "(* begin ARMC export *)\n"
    out = out + "let sigma_str =\n"
    out = out + "    \"s:0 "
    # Build Sigma set
    Sigma = set([])
    for d in M[1]:
        Sigma.add(d[1][0])
        Sigma.add(d[1][1])
    for s in Sigma:
        out = out + "%s:1 " % sanitize(s)
    out = out + "\"\n"
    out = out + "(* end ARMC export *)\n"
    return out

def exportInitARMC(aTSs):
    out = "(* begin ARMC export *)\n"
    out = out + "let init_str = \"\n"
    out = out + "    States p\n"
    out = out + "    Final States p\n"
    out = out + "    Transitions\n"
    out = out + "        s->p "
    # Initial states of transition systems
    I = [ TS[2] for TS in aTSs ]

    # Renaming of states
    Inew = []
    for i in I:
        inew = q2str(i)
        Inew = Inew + [ inew ]

    for i in Inew: 
        out = out + "%s(p)->p " % sanitize(i)
    out = out + "\"\n"
    out = out + "(* end ARMC export *)\n"
    return out

def exportTauARMC(M):
    out = "(* begin ARMC export *)\n"
    out = out + "let tau_str = (\n"
    # States
    out = out + "[ \"%s\"; " % sanitize(TAU_START_STATE)
    for q in M[0]:
        out = out + "\"%s\"; " % sanitize("_".join(q))
    out = out + "],\n"
    # Final States
    out = out + "[ \"%s\"; " % sanitize(TAU_START_STATE)
    for q in M[0]:
        out = out + "\"%s\"; " % sanitize("_".join(q))
    out = out + "],\n"
    # Transitions
    out = out + "[\n"
    out = out + "(\"s\", [], \"s\", \"%s\");\n" % sanitize(TAU_START_STATE)
    for q in M[0]:
        out = out + "(\"%s\", [\"%s\"], \"%s\", \"%s\");\n" % \
            (sanitize(q[0]), sanitize(TAU_START_STATE), sanitize(q[1]), sanitize("_".join(q)))
    for d in M[1]:
        out = out + "(\"%s\", [\"%s\"], \"%s\", \"%s\");\n" % \
            (sanitize(d[1][0]), sanitize("_".join(d[0])), sanitize(d[1][1]), sanitize("_".join(d[1])))
    out = out + "])\n"
    out = out + "(* end ARMC export *)\n"

    return out

def exportBadARMC(M, aBads):
    '''
    Generates ARMC string representing bad patterns. Parameters are:
        M = (Q, Delta)            Transducer representing interaction
                                  of instructions in the verified processor.
        aBads = [ bt1, ..., btN ] List of bad traces.
    '''

    out = "(* begin ARMC export *)\n"
    out = out + "let bad_str = \"\n"
    out = out + "    States "
    for i in range(len(aBads)):
        for j in range(len(aBads[i]) + 1):
            out = out + "p%d%d " % (i, j)
    out = out + "\n"
    out = out + "    Final States "
    for i in range(len(aBads)):
        if len(aBads[i]) > 0:
            out = out + "p%d%d " % (i, len(aBads[i]))
    out = out + "\n"
    out = out + "    Transitions\n"
    out = out + "        "
    for i in range(len(aBads)):
        out = out + "s->p%d0 " % i
    out = out + "\n"
    out = out + "\n"

    # Build Sigma set
    Sigma = set([])
    for d in M[1]:
        Sigma.add(d[1][0])
        Sigma.add(d[1][1])

    for i in range(len(aBads)):
        out = out + "        "
        qlast = "p%d0" % i
        for j in range(len(aBads[i])):
            q = "p%d%d" % (i, j + 1)
            out = out + "%s(%s)->%s " % (sanitize(aBads[i][j]), qlast, q)
    
            # Copy of Sigma 
            S = list(Sigma)
            # Remove curretly processed item
            S.remove(aBads[i][j])
            # Add self-loops to all remaining item in the set S
            for s in S:
                out = out + "%s(%s)->%s " % (sanitize(s), qlast, qlast) 
            out = out + "\n"
            out = out + "        "
            qlast = q

        out = out + "\n"
    out = out + "\"\n"
    out = out + "(* end ARMC export *)\n"
    return out

def exportSigmaStarARMC(M):
    out = "(* begin ARMC export *)\n"
    out = out + "let sigstar_str = \"\n"
    out = out + "    States p\n"
    out = out + "    Final States p\n"
    out = out + "    Transitions\n"
    out = out + "        s->p "
    # Build Sigma set
    Sigma = set([])
    for d in M[1]:
        Sigma.add(d[1][0])
        Sigma.add(d[1][1])
    for s in Sigma:
        out = out + "%s(p)->p " % sanitize(s)
    out = out + "\"\n"
    out = out + "(* end ARMC export *)\n"
    return out

def printTransducer(M):
    info("=== Transducer ===")
    info("=== States ===")
    for q in M[0]:
        info(str(q))
    info("=== Init States ===")
    for q in M[2]:
        info(str(q))
    info("=== Transitions ===")
    for t in M[1]:
        info(str(t[0]) + " --> " + str(t[1]))

def dotTransducer(M):
    # Header
    txt = "digraph transducer\n{\n"

    # Init state
    txt += sanitize(TAU_START_STATE) + " [ shape=box ];\n\n"

    # Transitions
    for q in M[0]:
        txt += "%s -> %s [ label=\"%s/%s\" ];\n" % \
            (sanitize(TAU_START_STATE), sanitize("_".join(q)), sanitize(q[0]), sanitize(q[1]))
    txt += "\n"
    for (q1, q2) in M[1]:
        txt += "%s -> %s [ label=\"%s/%s\" ];\n" % \
            (sanitize("_".join(q1)), sanitize("_".join(q2)), sanitize(q2[0]), sanitize(q2[1]))

    # Footer
    txt += "}\n"

    # Actual write
    fout = file("transducer.dot", "w")
    fout.write(txt)
    fout.close()

def printExecutionConditions(aFile, aPCs):
    info("Execution conditions:")
    info("    {0:16} {1:16} {2:8} {3:32}".format("Signal", "Value", "Phase", "Comment"))
    info("    " + "-" * 71)
    for pc in aPCs:
        comm = getComment(aFile, pc[0].name(), TSIG)
        if len(comm) > 32:
            comm = "... " + comm[-28:]
        info("    {0:16} {1:16} {2:8} {3:32}".format(pc[0].name(), pc[0].val(), \
            str(pc[1]), comm))

def stallPrune(aFile, aPhaseDict, aReps, aUarchs, aDefs, aUAsserts, aTS1, aTS2):
    (Q1, D1, qin1, C1, U1) = aTS1
    (Q2, D2, qin2, C2, U2) = aTS2
    path1 = Q1[1:-1]
    path2 = Q2[1:-1]

    # For each pair of transitions from TS1 and state TS2
    for q1 in path1:
        for q2 in path2:
            if debug_raw:
                info("Phase in (TS1, TS2): (%d, %d)" % (q1[2], q2[2]))

            # Skip lower phases, they are eliminated by construction of TS.
            if q1[1] >= q2[1]:
                continue

            # Current phase
            phase = q1[1]
            rep = aReps.get(phase, None)
            if rep == None:
                continue

            # Find stall signal of representative register
            st_sig = getCtrl(aFile, rep, CST)
            stall_zero = ConstCond(st_sig, 0, 1, "<-- can be stall zero")

            # Perform unrolls of transitions relation for both TS
            empty1  = { state:[] for state in path1 }
            upc1    = U1.shiftOrigin(-path1.index(q1))
            upc_st1 = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, \
                [ stall_zero.obj() ], path1, q1, empty1, COI_BCK, q1[0])
            upc1    = upc1.merge(upc_st1)

            #upc1 = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, \
            #    [ stall_zero.obj() ], path1, q1, C1, COI_BCK, q1[0])

            empty2  = { state:[] for state in path2 }
            upc2    = U2.shiftOrigin(-path2.index(q2))
            upc_st2 = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, \
                [ stall_zero.obj() ], path2, q2, empty2, COI_BCK, q2[0])
            upc2    = upc2.merge(upc_st2)

            #upc2 = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, \
            #    [ stall_zero.obj() ], path2, q2, C2, COI_BCK, q2[0])

            out = upc2smt(aFile, aPhaseDict, [ upc1, upc2 ], aUarchs, aUAsserts)
            out += stall_zero.smt() + "\n\n"

            # Run SMT solver and check whether the instance is SAT
            res = smtsat(out)

            if res == SMTSAT:
                if debug_raw:
                    info(smtrun2str(res))
                else:
                    if debug_dots:
                        sys.stdout.write(".")
                        sys.stdout.flush()
         
            if res == SMTUNSAT:
                if debug_raw:
                    info(smtrun2str(res))
                    info("Cond1: %s %s, Cond2: %s %s" % (q1[1], q1[2], q2[1], q2[2]))
                    info("q1: %s%s, q2: %s%s" % (q1[0], q1[2], q2[0], q2[2]))
                else:
                    if debug_dots:
                        sys.stdout.write("+")
                        sys.stdout.flush()

                # Successors of q2
                q2_next = [ t[2] for t in D2 if t[0] == q2 ]

                for q in q2_next:
                    if q != q2:
                        for t in D1:
                            if t[0] == q1 and t[2] != q1:
                                if q in t[1]:
                                    if debug_raw:
                                        info("Removing %s" % str(q))
                                    t[1].remove(q)
                    
            if res == SMTERR:
                error(smtrun2str(res))
                exit()

            #raw_input("Break...")
            #if q1[0] == "rd" and q1[1] == 2 and q2[0] == "wr" and q2[1] == 3:
            #    raw_input("Break...")

    if not debug_raw and debug_dots:
        print

    return aTS1

def clearPrune(aFile, aPhaseDict, aReps, aUarchs, aDefs, aUAsserts, aTS1, aTS2, aPath1, aPath2):
    (Q1, D1, qin1, C1, U1) = aTS1
    (Q2, D2, qin2, C2, U2) = aTS2

    for q1 in aPath1:
        if not q1 in Q1:
            error("A state of a path is not part of transition system!")

    for q2 in aPath2:
        if not q2 in Q2:
            error("A state of a path is not part of transition system!")

    # For each pair of transitions from TS1 and state TS2
    for q1 in aPath1:
        for q2 in aPath2:
            if debug_raw:
                info("Phase in (TS1, TS2): (%d, %d)" % (q1[2], q2[2]))

            # Skip lower phases, they are eliminated by construction of TS.
            if q1[1] >= q2[1]:
                continue

            # Current phase
            phase = q1[1]
            # We want clear signal of representative in the next phase,
            # since this a way the instruction are cleared
            rep = aReps.get(phase + 1, None)
            if rep == None:
                continue

            # Find stall signal of representative register
            cl_sig = getCtrl(aFile, rep, CCL)
            clear_zero = ConstCond(cl_sig, 0, 1, "<-- can clear be zero?")

            # Perform unrolls of transitions relation for both TS
            empty1  = { state:[] for state in aPath1 }
            upc1    = U1.shiftOrigin(-aPath1.index(q1))
            upc_cl1 = UnrolledPhaseConds.fromfile(aFile, aPhaseDict,
                [ clear_zero.obj() ], aPath1, q1, empty1, COI_BCK, q1[0])
            upc1    = upc1.merge(upc_cl1)

            empty2  = { state:[] for state in aPath2 }
            upc2    = U2.shiftOrigin(-aPath2.index(q2))
            upc_cl2 = UnrolledPhaseConds.fromfile(aFile, aPhaseDict,
                [ clear_zero.obj() ], aPath2, q2, empty2, COI_BCK, q2[0])
            upc2    = upc2.merge(upc_cl2)

            out = upc2smt(aFile, aPhaseDict, [ upc1, upc2 ], aUarchs, aUAsserts)
            out += clear_zero.smt() + "\n\n"

            # Run SMT solver and check whether the instance is SAT
            res = smtsat(out)

            if res == SMTSAT:
                if debug_raw:
                    info(smtrun2str(res))
                else:
                    if debug_dots:
                        sys.stdout.write(".")
                        sys.stdout.flush()

            if res == SMTUNSAT:
                if debug_raw:
                    info(smtrun2str(res))
                    info("Cond1: %s %s, Cond2: %s %s" % (q1[1], q1[2], q2[1], q2[2]))
                    info("q1: %s%s, q2: %s%s" % (q1[0], q1[2], q2[0], q2[2]))
                else:
                    if debug_dots:
                        sys.stdout.write("+")
                        sys.stdout.flush()

                # Successors of q2
                q2_post = [ t[2] for t in D2 if t[0] == q2 ]

                for q in q2_post:
                    if q != q2:
                        for t in D1:
                            if t[0] == q1 and t[2] != q1:
                                if q in t[1]:
                                    #info("Removing %s" % str(q))
                                    t[1].remove(q)

                for (mark_cl, phase_cl, flags_cl) in Q1:
                    if mark_cl == "cl" and phase_cl == q1[1] + 1:
                        #info("Adding transition to clear %s" % str((q1, q2_post, (mark_cl, phase_cl, flags_cl))))
                        D1.append((q1, q2_post, (mark_cl, phase_cl, flags_cl)))

            if res == SMTERR:
                error(smtrun2str(res))
                exit()

            #info("q1: %s, q2: %s" % (q2str(q1), q2str(q2)))
            #raw_input("Break...")
            #raw_input("Break...")
            #if q1[0] == "rd" and q1[1] == 2 and q2[0] == "wr" and q2[1] == 3:
            #    raw_input("Break...")

    if not debug_raw and debug_dots:
        print

    return (aTS1, aTS2)

#def stallPrune(aFile, aReps, aUarchs, aDefs, aUAsserts, aTS1, aTS2):
#    (Q1, D1, qin1, C1, R1, U1) = aTS1
#    (Q2, D2, qin2, C2, R2, U2) = aTS2
#
#    # For each pair of transitions from TS1 and state TS2
#    for q1 in Q1:
#        for q2 in Q2:
#            if debug_raw:
#                info("Phase in (TS1, TS2): (%d, %d)" % (q1[2], q2[2]))
#
#            # Skip lower phases, they are eliminated by construction of TS.
#            if q1[1] >= q2[1]:
#                continue
#
#            # Distance between stages
#            dist = q2[1] - q1[1]
#
#            # Successors of q2
#            q2_next = [ t[2] for t in D2 if t[0] == q2 ]
#
#            # Current phase
#            phase = q1[1]
#            rep = aReps.get(phase, None)
#            if rep == None:
#                continue
#
#            state_vars = []
#            state_assigns = []
#
#            for cond in C1[q1]:
#                state_vars = state_vars + [ cond[0] ] 
#                state_assigns = state_assigns + [ "(assert (= %s %s)) ; q1" % (cond[0], cond[1]) ]
#
#            for cond in C2[q2]:
#                state_vars = state_vars + [ cond[0] ] 
#                state_assigns = state_assigns + [ "(assert (= %s %s)) ; q2" % (cond[0], cond[1]) ]
#
#            for reg in R1[q1]:
#                state_vars = state_vars + [ getName(reg) ]
#
#            for reg in R2[q2]:
#                state_vars = state_vars + [ getName(reg) ]
#
#            st = getCtrl(aFile, rep, CST)
#            state_assigns = state_assigns + [ "(assert (= %s (_ 0 1))) ; st" % st[0] ]
#
#            state_vars = list(set(state_vars))
#            state_assigns = list(set(state_assigns))
#
#            sigs = ""
#            cmd = "(union "
#            for svar in state_vars:
#                cmd  += "(expr-coib 0 (select signal %s)) " % svar
#                sigs += "(select signal %s) " % svar
#
#            cmd += "(expr-coib 0 (stall (select register %s))) " % getName(rep)
#            cmd += ")"
#
#            # Get definition of relevant signals
#            # At first, collect all uarch regs
#            mclause = ""
#            for micro in aUarchs:
#                mclause = mclause + "(data-out (select register %s)) " % str(micro[0])
# 
#            # Then, merge them with previously extracted expression
#            r = simpleQuery(aFile, "(smtlib (union (references %s) %s %s))" % (cmd, mclause, sigs))
#            out = parseSmtlibExport(r) + "\n"
#
#            # Append definition of additional custom signal (e.g., symbolic address, ...)
#            for defi in aDefs:
#                out += smtlibExport("(declare-fun %s () (_ BitVec %d))\n" % defi)
#
#            # Get expressions that holds for given signals
#            r = simpleQuery(aFile, "(smtlib %s)" % cmd)
#            out += parseSmtlibExport(r) + "\n"
#
#            # Add optionally unrolled transition relation
#            out += U1[q1] + U2[q2] + "\n"
#
#            # Build assertions for SMTLib
#            for a in state_assigns:
#                out += re.sub("_ ", "_ bv", a) + "\n"
#
#            # Add assertions about microarch. regs
#            out += aUAsserts
#
#            # Add additional commands for SMTLib
#            out += "; Pre-printed command for possible debugging\n"
#            out += ";(apply ctx-solver-simplify) ; simplifies whole instance (Z3 only)\n"
#            out += "(check-sat)\n(exit)\n"
#
#            # Run SMT solver and collect result
#            res = runsmt(out)
#            if res == SMTSAT:
#                if debug_raw:
#                    info(smtrun2str(res))
#                else:
#                    if debug_dots:
#                        sys.stdout.write(".")
#                        sys.stdout.flush()
#         
#            if res == SMTUNSAT:
#                if debug_raw:
#                    info(smtrun2str(res))
#                    info("Cond1: %s %s, Cond2: %s %s" % (q1[1], q1[2], q2[1], q2[2]))
#                    info("q1: %s%s, q2: %s%s" % (q1[0], q1[2], q2[0], q2[2]))
#                else:
#                    if debug_dots:
#                        sys.stdout.write("+")
#                        sys.stdout.flush()
#
#                for q in q2_next:
#                    if q != q2:
#                        for t in D1:
#                            if t[0] == q1 and t[2] != q1:
#                                if q in t[1]:
#                                    if debug_raw:
#                                        info("Removing %s" % str(q))
#                                    t[1].remove(q)
#                    
#            if res == SMTERR:
#                error(smtrun2str(res))
#                exit()
#
#            #raw_input("Break...")
#            #if q1[0] == "rd" and q1[1] == 2 and q2[0] == "wr" and q2[1] == 3:
#            #    raw_input("Break...")
#                
#            #if q1[0] == "rd" and q1[2] == 2 and q2[0] == "wr" and q2[2] == 4:
#            #    raw_input("Break...")
#
#            #if q1[0] == "rd" and q1[2] == 2 and q2[0] == "wr" and q2[2] == 5:
#            #    raw_input("Break...")
#
#            #if q1[0] == "rd" and q1[2] == 5 and q2[0] == "wr" and q2[2] == 5:
#            #    raw_input("Break...")
#
#            #print q1[0], q1[1], q2[0], q2[1]
#            #pause()
#            #if q1[0][:2] == "rd" and q1[1] == 2 and q2[0][:2] == "wr" and q2[1] == 3:
#            #    pause()
#
#    if not debug_raw and debug_dots:
#        print
#
#    return aTS1

def checkConf(aFile, aPhaseDict, aUarchs, aDefs, aUAsserts, aTS):
    (Q1, D1, qin1, C1, U1) = aTS

    # TODO: For any state of TS
    #q1 = Q1[2]

    #path = Q1[1:-1]
    #upc = UnrolledPhaseConds.fromfile(aFile, aPhaseDict, [], \
    #    path, q1, C1, COI_BCK, "")
    out = upc2smt(aFile, aPhaseDict, [ U1 ], aUarchs, aUAsserts)

    # Run SMT solver and check whether the instance is SAT
    res = smtsat(out)
    if res == SMTSAT:
        return True
   
    if res == SMTUNSAT:
        return False
            
    if res == SMTERR:
        error(smtrun2str(res))
        exit()
        return False

#def checkConf(aFile, aUarchs, aDefs, aUAsserts, aTS):
#    (Q1, D1, qin1, C1, R1, U1) = aTS
#
#    # For any stat of TS
#    q1 = Q1[2]   
# 
#    state_vars = []
#    state_assigns = []
#
#    for cond in C1[q1]:
#        state_vars = state_vars + [ cond[0] ] 
#        state_assigns = state_assigns + [ "(assert (= %s %s)) ; q1" % (cond[0], cond[1]) ]
#
#    for reg in R1[q1]:
#        state_vars = state_vars + [ getName(reg) ]
#
#    state_vars = list(set(state_vars))
#    state_assigns = list(set(state_assigns))
#
#    sigs = ""
#    cmd = "(union "
#    for svar in state_vars:
#        cmd  += "(expr-coib 0 (select signal %s)) " % svar
#        sigs += "(select signal %s) " % svar
#
#    cmd += ")"
#
#    # Get definition of relevant signals
#    # At first, collect all uarch regs
#    mclause = ""
#    for micro in aUarchs:
#        mclause = mclause + "(data-out (select register %s)) " % str(micro[0])
# 
#    # Then, merge them with previously extracted expression
#    r = simpleQuery(aFile, "(smtlib (union (references %s) %s %s))" % (cmd, mclause, sigs))
#    out = parseSmtlibExport(r) + "\n"
#
#    # Append definition of additional custom signal (e.g., symbolic address, ...)
#    for defi in aDefs:
#        out += smtlibExport("(declare-fun %s () (_ BitVec %d))\n" % defi)
# 
#    # Get expressions that holds for given signals
#    r = simpleQuery(aFile, "(smtlib %s)" % cmd)
#    out += parseSmtlibExport(r) + "\n"
#
#    # Add optionally unrolled transition relation
#    out += U1[q1] + "\n"
#
#    # Build assertions for SMTLib
#    for a in state_assigns:
#        out += re.sub("_ ", "_ bv", a) + "\n"
#
#    # Add assertions about microarch. regs
#    out += aUAsserts
#
#    # Add addtional commands for SMTLib
#    out += "(check-sat)\n(exit)\n"
#
#    # Run SMT solver and collect result
#    res = runsmt(out)
#    if res == SMTSAT:
#        return True
#   
#    if res == SMTUNSAT:
#        return False
#            
#    if res == SMTERR:
#        error(smtrun2str(res))
#        exit()
#        return False

def inspectRAW(aFile, aPhases, aWrites, aReads, aPhaseDict, aReps, aUarchs, aUAsserts, aDsts, raw, aStat):
    '''
    Inspects a possibility of RAW hazard for one problematic part of
    a processor.
    '''

    # We must assume both register and memories. For the later ones,
    # we have to determine and analyse all read ports as possible
    # sources and destinations.
    sws = [] # Written storages
    srs = [] # Read storages
    if getType(raw) == TMEMORY:
        for sw in getWritePorts(aFile, raw):
            sws += [ sw ]
        for sr in getReadPorts(aFile, raw):
            srs += [ sr ]
    elif getType(raw) == TREG:
        sws += [ raw ]
        srs += [ raw ]
    else:
        error("Unexpected type of RTL object %s" % str(raw))

    # Items that cannot be read or written cannot cause RAW hazards.
    if len(sws) < 1 or len(srs) < 1:
        return

    pipe_len = max(aReps)

    info("****************************************************************************")
    info("*** Inspection of possible RAW hazards on %s" % getName(raw))
    info("****************************************************************************")

    # Sets of possible hazard cases
    forward_hazard_cases = []
    turn_hazard_cases = []

    # For each readable storage W and its write phase
    for sw in sws:
        for pw in getWrites(aWrites, sw):
            # For each readable storage R and its read phase
            for sr in srs:
                for pr in getReads(aReads, sr):
                    # Throw away the case of RAW hazard within instruction
                    # itself. This is already checked during verification of
                    # instructions in isolation.
                    if (pw - pr) <= 1:
                        continue

                    for st in aDsts:
                        for pt in getWrites(aWrites, st[0]):
                            # Destination T can be only written after read of the source R,
                            # otherwise it means that there is RAW hazard on the source R
                            # which we will necessary cover in our set of hazard cases.
                            if pt < pr:
                                continue

                            info("Investigating time variant:")
                            info("    Storage (v_w): %s written at (t_w): %d" % \
                                (str(sw), pw))
                            info("    Storage (v_r): %s read    at (t_r): %d" % \
                                (str(sr), pr))
                            info("    Storage (v_p): %s written at (t_p): %d on "
                                "connection: %s" %\
                                (str(st[0]), pt, st[1]))

                            # We evaluate all possible paths from the problematic part that
                            # end in writing to some other part.
                            paths = extractPaths(aFile, sr, st[0], st[1])
                            
                            info("Building a set of forward hazard cases ...")
                            fhcs = getForwardHazardCases(aFile, paths, \
                                pr, pt, aWrites, aReads)

                            for fhc in fhcs:
                                new_case = (sw, pw, sr, pr, st, pt, fhc[0], fhc[1])
                                if not new_case in forward_hazard_cases:
                                    forward_hazard_cases += [ new_case ]

                            #info("Building a set of turning hazard cases ...")
                            #thcs = getTurnHazardCases(aFile, paths, pr, pt, aWrites, aReads)
                            #for thc in thcs:
                            #    new_case = (sw, pw, sr, pr, thc[0], thc[1], thc[2], thc[3])
                            #    if not new_case in turn_hazard_cases:
                            #        turn_hazard_cases += [ new_case ]


    if debug_dots:
        print
 
    #for fhc in forward_hazard_cases:
    #    printRAWHC(aFile, fhc)

    #for thc in turn_hazard_cases:
    #    printRAWHC(aFile, thc)

    #for thc in turn_hazard_cases:
    #    included = False
    #    for fhc in forward_hazard_cases:
    #        included = False
    #        if len(fhc[7]) < len(thc[7]):
    #            continue

    #        included = True
    #        for i in range(len(thc[7])):
    #            if thc[7][i] != fhc[7][i]:
    #                included = False
    #                break

    #        if included:
    #            break
    #        
    #    if not included:
    #        warning("Turn case path %s not included anywhere" % str(thc[7]))
    #        pause()

    
    thc_dst_phases = [ thc[5] for thc in turn_hazard_cases ]
    if thc_dst_phases != []:
        thc_dst_phase = min(thc_dst_phases) + 1
    else:
        thc_dst_phase = None

    # Perform deeper analysis for each hazard case
    cnt = 0
    total = len(forward_hazard_cases)
    # TODO: include turn hazard cases ...
    hFile = open(aStat, "a")
    asserts = hFile.write("RAW;%s;%d\n" % (getType(raw) + "_" + getName(raw), total))
    hFile.close()
    for (sw, pw, sr, pr, st, pt, path_conds, path_objs) in forward_hazard_cases:
        cnt += 1
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        info("Hazard case %d / %d" % (cnt, total))
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        printRAWHC(aFile, (sw, pw, sr, pr, st, pt, path_conds, path_objs))
        if thc_dst_phase != None and pr < thc_dst_phase:
            info("Destination phase override by turn case to: %d" % thc_dst_phase)
            hazard_len = thc_dst_phase - pr
        else:
            hazard_len = pt - pr

        if thc_dst_phase != None and pr < thc_dst_phase:
            r_mark_phase = thc_dst_phase
        else:
            r_mark_phase = pt

        # Get relevant access conditions to the elements of the currently
        # processed hazard.
        info("Extracting execution conditions ...")
        sw_condss = extractWriteConditions(aFile, aPhaseDict, aUAsserts, \
            aUarchs, aWrites, sw, TD, "1", "wr", pw)
        sr_condss = extractReadConditions(aFile, sr, "1", "rd", pr)
        st_condss = extractWriteConditions(aFile, aPhaseDict, aUAsserts, \
            aUarchs, aWrites, st[0], st[1], "2", "rd", pt)

        # TODO: ...
        #if dst[1] == CCL:
        #    # FIXME: Not tested ...
        #    rep = aReps.get(dst_phase - 1, None)
        #    if rep != None:
        #        # There must not be a stall at the previous phase during
        #        # a clear of pipeline register.
        #        st = getCtrl(aFile, rep, CST)
        #        write_conds += [ ((getName(st), "(_ 0 1)"), dst_phase - 1) ]

        #        en = getCtrl(aFile, rep, CEN)
        #        write_conds += [ ((getName(en), "(_ 1 1)"), dst_phase - 1) ]

        # Additional definitions, i.e., used symbolic addresses
        defs = []
        if getType(sr) == TMEM:
            defs += [ (SYM_ADDRESS_NAME + "1", getAddressSize(aFile, sr)) ]
        if getType(st[0]) == TMEM:
            defs += [ (SYM_ADDRESS_NAME + "2", getAddressSize(aFile, st[0])) ]

        # Show path conditions
        info("Path conditions ...")
        printExecutionConditions(aFile, path_conds)

        # Add unique marker to path conditions
        mp_conds = markPathConditions(path_conds, "rd")

        # For each combination of the previously extracted conditions
        for sw_conds in sw_condss:
          for sr_conds in sr_condss:
             for st_conds in st_condss:
                # Merge (i) conditions for reading from R, (ii) marked path
                # conditions, and (iii) conditions for writing to T in order
                # to form a reading trace.
                rd_conds = sr_conds + mp_conds + st_conds

                # Build transition systems 
                info("Building transition systems ...")
                # Transition system for writing instruction
                wr_trace = buildBasicTS(aFile, aPhaseDict, sw_conds, "wr", 1, \
                    pipe_len, pw, pw + hazard_len, -1, False)
                # Transition system for reading instruction
                rd_trace_len = max(pipe_len, r_mark_phase)
                rd_trace = buildBasicTS(aFile, aPhaseDict, rd_conds, "rd", 1, \
                    rd_trace_len, -1, -1, r_mark_phase, False)
                # Transition system for filling instruction
                fi_trace = buildBasicTS(aFile, aPhaseDict, [], "fi", 1, \
                    2, -1, -1, -1, False)
        
                info("Checking for inner conflicts ...")
                if not checkConf(aFile, aPhaseDict, aUarchs, defs, \
                    aUAsserts, rd_trace):
                    info("Execution conditions spurious ...")
                    continue
        
                # Add guards from other systems
                info("Finalizing guards ...")
                completeGuards(wr_trace, rd_trace)
                #completeGuards(wr_trace, fi_trace)
                #completeGuards(rd_trace, fi_trace)
        
                # Prune behaviour of TS according because of possible stalls
                # and conflicts.
                info("Pruning transition relations of transition systems ...")
                rd_trace = stallPrune(aFile, aPhaseDict, aReps, aUarchs, defs, \
                    aUAsserts, rd_trace, wr_trace)

                info("Write trace:")
                printTS(wr_trace)
        
                info("Read trace:")
                printTS(rd_trace)

                #info("Fill trace:")
                #printTS(fi_trace)
                
                info("Building transducer for ARMC ...")
                M = buildTransducer([wr_trace, rd_trace, fi_trace])
                #printTransducer(M)
        
                # Construct "BAD" traces
                Bad = []
                for qwr in wr_trace[0]:
                    for qrd in rd_trace[0]:
                        if "h" in qwr[2] and "r" in qrd[2]:
                            Bad += [ [q2str(qwr), q2str(qrd)] ]
        
                print Bad
        
                info("Exporting problem to ARMC ...")
                con = headerARMC(aFile)
                con += "let rtl _ =\n"
                con += exportSigmaARMC(M)
                con += "in\n"
                con += exportInitARMC([wr_trace, rd_trace, fi_trace])
                con += "in\n"
                con += exportTauARMC(M)
                con += "in\n"
                con += exportBadARMC(M, Bad)
                con += "in\n"
                con += exportSigmaStarARMC(M)
                con += "in\n"
                con += footerARMC()
        
                saveARMC(ARMC_EXPORT_FILE, con)
                info("ARMC File created: %s" % ARMC_EXPORT_FILE)
        
                info("Running ARMC ...")
                armc_run = runarmc(ARMC_EXPORT_FILE)
                if armc_run == ARMCHOLD:
                    info("-" * 76)
                    info("--- ARMC Result: %s" % armcrun2str(armc_run))
                    info("-" * 76)
                else:
                    error("!" * 76)
                    error("!!! ARMC Result: %s" % armcrun2str(armc_run))
                    error("!!! Path %s --> %s (%s)" % (sr, st[0], st[1]))
                    error("!" * 76)
                    if stop_error:
                        raw_input("Error found! Press Enter to continue...")
        
                if debug_raw:
                    if debug_stop:
                        raw_input("Press Enter to continue...")
 
def inspectWAW(aFile, aPhases, aWrites, aReads, aPhaseDict, aReps, aUarchs, \
    aUAsserts, waw, aStat):
    '''
    Inspects a possibility of WAW hazard for one problematic part of
    a processor.
    '''

    # We must assume both register and memories. For the later ones,
    # we have to determine and analyse all read ports as possible
    # destinations.
    sws = [] # Written storages
    if getType(waw) == TMEMORY:
        for sw in getWritePorts(aFile, waw):
            sws += [ sw ]
    elif getType(waw) == TREG:
        sws += [ waw ]
    else:
        error("Unexpected type of RTL object %s" % str(waw))

    # Items that cannot be written cannot cause WAW hazards.
    if len(sws) < 1:
        return

    # Length of the pipeline
    pipe_len = max(aReps)

    info("****************************************************************************")
    info("*** Inspection of possible WAW hazards on %s" % getName(waw))
    info("****************************************************************************")

    # Set of possible write phases for each destination
    write_phases = []
    for sw in sws:
        for ph in getWrites(aWrites, sw):
            write_phases += [ (sw, ph) ] 

    # Set of possible hazard cases
    hazard_cases = [] 

    # Extract hazard cases for the problematic resources
    for (dst1, ph1) in write_phases:
        for (dst2, ph2) in write_phases:
            if ph1 < ph2:
                hazard_cases += [ (dst1, ph1, dst2, ph2) ]

    # For each hazard case
    cnt = 0
    total = len(hazard_cases)
    hFile = open(aStat, "a")
    asserts = hFile.write("WAW;%s;%d\n" % (getType(waw) + "_" + getName(waw), total))
    hFile.close()
    for (dst1, ph1, dst2, ph2) in hazard_cases:
        cnt += 1
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        info("Hazard case %d / %d" % (cnt, total))
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        printWAWHC(aFile, (dst1, ph1, dst2, ph2))

        # Determine length of the hazard
        hazard_len = ph2 - ph1
        # Phase when later instruction writes to its target storage
        r_mark_phase = ph2

        info("Extracting execution conditions ...")
        # Get relevant write conditions that must hold when writing
        # to a problematic part (register or write port) of the processor.
        wr_conds = {}
        for (dst, ph) in [ (dst1, ph1), (dst2, ph2) ]:
            wr_conds[(dst, ph)] = extractWriteConditions(aFile, aPhaseDict,
                aUAsserts, aUarchs, aWrites, dst, TD, "", "wr%d" % ph, ph)

        # For each combination of write conditions
        for wr_conds1 in wr_conds[(dst1, ph1)]:
            for wr_conds2 in wr_conds[(dst2, ph2)]:
                # Additional definitions
                defs = []
                if getType(dst1) == TMEM:
                    defs += [ (SYM_ADDRESS_NAME, getAddressSize(aFile, dst1)) ]

                info("Building transition systems ...")
                wr_trace1 = buildBasicTS(aFile, aPhaseDict, wr_conds1, "wr%d" % ph1,
                    1, pipe_len, ph1, ph1 + hazard_len, -1, False)
                wr_trace2_len = max(pipe_len, r_mark_phase)
                wr_trace2 = buildBasicTS(aFile, aPhaseDict, wr_conds2, "wr%d" % ph2,
                    1, wr_trace2_len, -1, -1, r_mark_phase, False)
                fi_trace = buildBasicTS(aFile, aPhaseDict, [], "fi", 1, 2, -1, -1, -1, False)

                info("Finalizing guards ...")
                # Add guards from other systems
                completeGuards(wr_trace1, wr_trace2)

                info("Pruning transition relations of transition systems ...")
                # Prune behaviour of TS according because of possible stalls and conflicts
                wr_trace1 = stallPrune(aFile, aPhaseDict, aReps, aUarchs, defs,
                    aUAsserts, wr_trace1, wr_trace2)

                # Show created traces
                info("Write trace 1:")
                printTS(wr_trace1)
                #info("Write trace 2:")
                #printTS(wr_trace2)

                # Create transducer for ARTMC
                info("Building transducer for ARMC ...")
                M = buildTransducer([wr_trace1, wr_trace2, fi_trace])

                # Construct "BAD" traces
                Bad = []
                for qwr1 in wr_trace1[0]:
                    for qwr2 in wr_trace2[0]:
                        if "h" in qwr1[2] and "r" in qwr2[2]:
                            Bad += [ [q2str(qwr2), q2str(qwr1)] ]

                info("Exporting problem to ARMC ...")
                con  = headerARMC(aFile)
                con += "let rtl _ =\n"
                con += exportSigmaARMC(M)
                con += "in\n"
                con += exportInitARMC([wr_trace1, wr_trace2, fi_trace])
                con += "in\n"
                con += exportTauARMC(M)
                con += "in\n"
                con += exportBadARMC(M, Bad)
                con += "in\n"
                con += exportSigmaStarARMC(M)
                con += "in\n"
                con += footerARMC()

                saveARMC(ARMC_EXPORT_FILE, con)
                info("ARMC File created: %s" % ARMC_EXPORT_FILE)

                info("Running ARMC ...")
                armc_run = runarmc(ARMC_EXPORT_FILE)
                if armc_run == ARMCHOLD:
                    info("-" * 76)
                    info("--- ARMC Result: %s" % armcrun2str(armc_run))
                    info("-" * 76)
                else:
                    error("!" * 76)
                    error("!!! ARMC Result: %s" % armcrun2str(armc_run))
                    error("!" * 76)
                    if stop_error:
                        raw_input("Error found! Press Enter to continue...")

                if debug_raw:
                    if debug_stop:
                        raw_input("Press Enter to continue...")

def inspectWAR(aFile, aPhases, aWrites, aReads, aPhaseDict, aReps, aUarchs, \
    aUAsserts, aDsts, war, aStat):
    '''
    Inspects a possibility of WAR hazard for one problematic part of
    a processor.
    '''

    # We must assume both register and memories. For the later ones,
    # we have to determine and analyse all read ports as possible
    # sources and destinations.
    sws = [] # Written storages
    srs = [] # Read storages
    if getType(war) == TMEMORY:
        for sw in getWritePorts(aFile, war):
            sws += [ sw ]
        for sr in getReadPorts(aFile, war):
            srs += [ sr ]
    elif getType(war) == TREG:
        sws += [ war ]
        srs += [ war ]
    else:
        error("Unexpected type of RTL object %s" % str(war))

    # Items that cannot be read or written cannot cause WAR hazards.
    if len(sws) < 1 or len(srs) < 1:
        return

    # Length of the pipeline
    pipe_len = max(aReps)

    info("****************************************************************************")
    info("*** Inspection of possible WAR hazards on %s" % getName(war))
    info("****************************************************************************")

    # Sets of possible hazard cases
    hazard_cases = []

    # For each readable storage R and its read phase
    for sr in srs:
        for pr in getReads(aReads, sr):
            # For each possible destination T and its write phase
            for st in aDsts:
                for pt in getWrites(aWrites, st[0]):
                    # Skip combinations where write to T precedes read from R,
                    # the WAR hazard cannot occur here.
                    if pt < pr:
                        continue

                    # For each writable storage W and its write phase
                    for sw in sws:
                        for pw in getWrites(aWrites, sw):
                            # Skip combination where read from R precedes write to W,
                            # the WAR hazard cannot occur.
                            if pr < pw:
                                continue

                            info("Investigating time variant:")
                            info("    Storage (v_w): %s written at (t_w): %d" % \
                                (str(sw), pw))
                            info("    Storage (v_r): %s read    at (t_r): %d" % \
                                (str(sr), pr))
                            info("    Storage (v_p): %s written at (t_p): %d on "
                                "connection: %s" % \
                                (str(st[0]), pt, st[1]))

                            # We evaluate all possible paths from the problematic part that
                            # end in writing to some other part.
                            paths = extractPaths(aFile, sr, st[0], st[1])
 
                            info("Building a set of forward hazard cases ...")
                            fhcs = getForwardHazardCases(aFile, paths, \
                                pr, pt, aWrites, aReads)

                            for fhc in fhcs:
                                new_case = (sr, pr, st, pt, sw, pw, fhc[0], fhc[1])
                                if not new_case in hazard_cases:
                                    hazard_cases += [ new_case ]

    if debug_dots:
        print
 
    # Perform deeper analysis for each hazard case
    cnt = 0
    total = len(hazard_cases)
    hFile = open(aStat, "a")
    asserts = hFile.write("WAR;%s;%d\n" % (getType(war) + "_" + getName(war), total))
    hFile.close()
    for (sr, pr, st, pt, sw, pw, path_conds, path_objs) in hazard_cases:
        cnt += 1
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        info("Hazard case %d / %d" % (cnt, total))
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        printWARHC(aFile, (sr, pr, st, pt, sw, pw, path_conds, path_objs))

        # Hazard length is understand as distance between writing to storage W
        # and reading from R.
        hazard_len = pr - pw

        # Phase where the data are committed to storage W
        r_mark_phase = pw

        # Get relevant access conditions to the elements of the currently
        # processed hazard.
        info("Extracting execution conditions ...")
        sw_condss = extractWriteConditions(aFile, aPhaseDict, aUAsserts, \
            aUarchs, aWrites, sw, TD, "1", "wr", pw)
        sr_condss = extractReadConditions(aFile, sr, "1", "rd", pr)
        st_condss = extractWriteConditions(aFile, aPhaseDict, aUAsserts, \
            aUarchs, aWrites, st[0], st[1], "2", "rd", pt)

        # Additional definitions, i.e., used symbolic addresses
        defs = []
        if getType(sr) == TMEM:
            defs += [ (SYM_ADDRESS_NAME + "1", getAddressSize(aFile, sr)) ]
        if getType(st[0]) == TMEM:
            defs += [ (SYM_ADDRESS_NAME + "2", getAddressSize(aFile, st[0])) ]

        # Show path conditions
        info("Path conditions ...")
        printExecutionConditions(aFile, path_conds)

        # Add unique marker to path conditions
        mp_conds = markPathConditions(path_conds, "rd")

        # For each combination of the previously extracted conditions
        for sw_conds in sw_condss:
            for sr_conds in sr_condss:
                for st_conds in st_condss:
                    # Merge (i) conditions for reading from R, (ii) marked path
                    # conditions, and (iii) conditions for writing to T in order
                    # to form a reading trace.
                    rd_conds = sr_conds + mp_conds + st_conds
               
                    # Build transition systems
                    info("Building transition systems ...")
                    wr_trace_len = max(pipe_len, r_mark_phase)
                    wr_trace = buildBasicTS(aFile, aPhaseDict, sw_conds, "wr", 1, \
                        wr_trace_len, -1, -1, r_mark_phase, False)
                    rd_trace = buildBasicTS(aFile, aPhaseDict, rd_conds, "rd", 1, \
                        pipe_len, pr-hazard_len+1, pr+1, -1, False)
                    fi_trace = buildBasicTS(aFile, aPhaseDict, [], "fi", 1, \
                        2, -1, -1, -1, False)

                    info("Checking for inner conflicts ...")
                    if not checkConf(aFile, aPhaseDict, aUarchs, defs, aUAsserts, rd_trace):
                        info("Execution conditions spurious ...")
                        continue

                    # Add guards from other systems
                    info("Finalizing guards ...")
                    completeGuards(wr_trace, rd_trace)
                    #completeGuards(wr_trace, fi_trace)
                    #completeGuards(rd_trace, fi_trace)

                    # Prune behaviour of TS according because of possible stalls
                    # and conflicts.
                    info("Pruning transition relations of transition systems ...")
                    wr_trace = stallPrune(aFile, aPhaseDict, aReps, aUarchs, defs, \
                        aUAsserts, wr_trace, rd_trace)

                    info("Write trace:")
                    printTS(wr_trace)

                    info("Read trace:")
                    printTS(rd_trace)

                    info("Building transducer for ARMC ...")
                    M = buildTransducer([wr_trace, rd_trace, fi_trace])
                    #printTransducer(M)
        
                    # Construct "BAD" traces
                    Bad = []
                    for qwr in wr_trace[0]:
                        for qrd in rd_trace[0]:
                            if "r" in qwr[2] and "h" in qrd[2]:
                                Bad += [ [q2str(qrd), q2str(qwr)] ]
        
                    #print Bad
        
                    info("Exporting problem to ARMC ...")
                    con = headerARMC(aFile)
                    con += "let rtl _ =\n"
                    con += exportSigmaARMC(M)
                    con += "in\n"
                    con += exportInitARMC([wr_trace, rd_trace, fi_trace])
                    con += "in\n"
                    con += exportTauARMC(M)
                    con += "in\n"
                    con += exportBadARMC(M, Bad)
                    con += "in\n"
                    con += exportSigmaStarARMC(M)
                    con += "in\n"
                    con += footerARMC()
        
                    saveARMC(ARMC_EXPORT_FILE, con)
                    info("ARMC File created: %s" % ARMC_EXPORT_FILE)
        
                    info("Running ARMC ...")
                    armc_run = runarmc(ARMC_EXPORT_FILE)
                    if armc_run == ARMCHOLD:
                        info("-" * 76)
                        info("--- ARMC Result: %s" % armcrun2str(armc_run))
                        info("-" * 76)
                    else:
                        error("!" * 76)
                        error("!!! ARMC Result: %s" % armcrun2str(armc_run))
                        error("!!! Path %s --> %s (%s)" % (sr, st[0], st[1]))
                        error("!" * 76)
                        if stop_error:
                            raw_input("Error found! Press Enter to continue...")
        
                    if debug_raw:
                        if debug_stop:
                            raw_input("Press Enter to continue...")

def getJmpCases(aFile, aPhaseDict, aObj, aPhase, aWrites, aReads):
    """
    Extracts edge conditions that necessary in order the given storage to be
    written during jump.
    """

    jcs = []

    # Determine write time
    # TODO: Generalize
    wt = 1

    # Extract input data to the register
    data_in = getDatain(aFile, aObj)
    if data_in is None:
        return []

    # Data-in edge of the program counter should be in the given stage
    if not aPhase-wt in aWrites[data_in]:
        return []

    phase_sigs = []
    for sig in getExprDown(aFile, data_in):
        wphases = aWrites[sig]
        # We are interested only in signals written by a single phase
        if (len(wphases) != 1):
            continue
        if (getType(sig) == TSIG) and (not 1 in wphases) and \
            (aPhase-wt in wphases) and (len(aReads[sig]) > 1 or not 0 in aReads[sig]):
            # print "jmp_cases sig: %s, %d in %s (%s)" % (getName(sig), aPhase-wt, str(wphases), getComment(aFile, getName(sig), getType(sig)))
            phase_sigs += [ sig ]

    # REM: Old implementation
    # phase_sigs = []
    # for (sig, wphases) in aWrites.items():
    #     if (getType(sig) == TSIG) and (not 1 in wphases) and \
    #             (aPhase-wt in wphases) and (len(aReads[sig]) > 1 or not 0 in aReads[sig]):
    #         paths = extractSigPaths(aFile, sig, data_in)
    #         if len(paths) > 0:
    #             print "jmp_cases sig: %s, %d in %s (%s)" % (getName(sig), aPhase-wt, str(wphases), getComment(aFile, getName(sig), getType(sig)))
    #             phase_sigs += [ sig ]

    # Computation of the COI-Down mapping st. COI-Down(s) = Expr-Coib(s)
    # for the previously calculated set of influencing signals.
    coi_downs = [ (sig, set(getExprDown(aFile, sig) + [sig])) for sig in phase_sigs ]

    # Find maximal elements among item of the downward mapping.
    max_elems = getMaximalElements(coi_downs,
        lambda (s1, d1), (s2, d2): d1.issubset(d2))

    # For our purpose, it is enough to use only the computed
    # maximal elements.
    phase_sigs = [ i[0] for i in max_elems ]

    pathss = [ extractSigPaths(aFile, sig, data_in) for sig in phase_sigs ]
    for paths in pathss:
        for path in paths:
            path_objs = extractPathObjects(path)
            path_valid = True

            for obj in path_objs:
                # Eliminate path containing registers
                # TODO: include memories ...
                if getType(obj) == TREG:
                    path_valid = False
                    if debug_dots:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                    break

            if path_valid and not (aPhase-wt in aWrites[obj]):
                # print "phase: %s not in %s" % (aPhase-wt, aWrites[aObj])
                path_valid = False
                if debug_dots:
                    sys.stdout.write("~")
                    sys.stdout.flush()
 
            if path_valid:
                path_conds = extractPathConditions(aFile, path)
                # Only paths with path conditions are relevant
                # TODO: double check ...
                if len(path_conds) == 0:
                    if debug_dots:
                        sys.stdout.write("-")
                        sys.stdout.flush()
                    continue

                path_conds = recountConditionPhases(aPhase-wt, path_conds)
                jcs += [ (path_conds, path_objs) ]
                if debug_dots:
                    sys.stdout.write("+")
                    sys.stdout.flush()

    if debug_dots:
        print

    return jcs

def inspectCTL(aFile, aPhases, aWrites, aReads, aPhaseDict, aReps, aUarchs,
    aUAsserts, aDsts, ctl, aStat):
    '''
    Inspects a possibility of control hazard for one problematic part of
    a processor.
    '''

    # Let's check correctness of arguments.
    sws = [] # Written storages
    if getType(ctl) == TREG:
        sws += [ ctl ]
    else:
        error("Unexpected type of RTL object %s" % str(ctl))

    # Items that cannot be written cannot cause CTL hazards.
    if len(sws) < 1:
        return

    # Length of the pipeline
    pipe_len = max(aReps)

    info("****************************************************************************")
    info("*** Inspection of possible CTL hazards on %s" % getName(ctl))
    info("****************************************************************************")

    # Set of possible hazard cases
    hazard_cases = []

    # Extract hazard cases for the problematic resources
    # For each writable storage W and its write phase
    for sw in sws:
        for pw in getWrites(aWrites, sw):
            # For each possible destination T
            for st in aDsts:
                # We interested in to enable connections only in this case.
                if st[1] != TD:
                    continue

                # For write phase of destination T
                for pt in getWrites(aWrites, st[0]):
                    info("Investigating time variant:")
                    info("    Storage (v_w): %s written at (t_w): %d" % \
                        (str(sw), pw))
                    info("    Storage (v_p): %s written at (t_p): %d on "
                        "connection: %s" %\
                        (str(st[0]), pt, st[1]))

                    # We evaluate all possible paths from the phases higher than one that
                    # end in writing to some program counter.
                    info("Building a set of hazard cases ...")
                    jcs = getJmpCases(aFile, aPhaseDict, sw, pw, aWrites, aReads)
                    for jc in jcs:
                        new_case = (sw, pw, st, pt, jc[0], jc[1])
                        if not new_case in jcs:
                            hazard_cases += [ new_case ]

    # For each hazard case
    cnt = 0
    total = len(hazard_cases)
    hFile = open(aStat, "a")
    asserts = hFile.write("CTL;%s;%d\n" % (getType(ctl) + "_" + getName(ctl), total))
    hFile.close()
    for (sw, pw, st, pt, path_conds, path_objs) in hazard_cases:
        cnt += 1
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        info("Hazard case %d / %d" % (cnt, total))
        info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        printCTLHC(aFile, (sw, pw, st, pt, path_conds, path_objs))

        # Distance between writing to storage W and writing to T.
        distance = pw - pt
        #print "dist", distance

        # Hazard length denotes the number of instruction that may
        # enter pipeline after jump instruction
        hazard_len = pw - 1
        #print "hl", hazard_len

        # No. of hazardous states on the size corresponding to write
        # to storage W
        hbr = hazard_len - max(distance, 1) + 1
        #print "hbr", hbr

        # No. of hazardous states on the size corresponding to write
        # to storage T
        hwr = max(distance, 1)
        #print "hwr", hwr

        info("Extracting execution conditions ...")
        # Get relevant write conditions that must hold when writing
        # to a problematic part (register or write port) of the processor.
        sw_condss = extractWriteConditions(aFile, aPhaseDict,
            aUAsserts, aUarchs, aWrites, sw, TD, "", "br", pw)

        # Show path conditions
        info("Jump conditions ...")
        printExecutionConditions(aFile, path_conds)

        # Get relevant write conditions that must hold when writing
        # to another storage of the processor.
        st_condss = extractWriteConditions(aFile, aPhaseDict, aUAsserts,
            aUarchs, aWrites, st[0], st[1], "", "wr", pt)

        # Additional definitions, i.e., used symbolic addresses
        defs = []
        if getType(st[0]) == TMEM:
            defs += [ (SYM_ADDRESS_NAME, getAddressSize(aFile, st[0])) ]

        # Add unique marker to path conditions
        mp_conds = markPathConditions(path_conds, "br")

        # For each combination of the previously extracted conditions
        for sw_conds in sw_condss:
            for st_conds in st_condss:
                # Combine corresponding path conditions
                br_conds = sw_conds + mp_conds
                wr_conds = st_conds

                info("Building transition systems ...")
                fi_trace = buildPlainTS(aFile, aPhaseDict, "fi", 1,
                    max(pipe_len + hwr - 1, pipe_len + hbr - 1))

                # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
                # A. Writing trace
                # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

                # 1. Build main and clear trace for writing instruction
                wr_main_trace = buildCondMappedTS(aFile, aPhaseDict, wr_conds,
                    "wr", 1, pipe_len + hwr - 1)
                wr_clr_trace = buildPlainTS(aFile, aPhaseDict,
                    "cl", 1, pipe_len + hwr - 1)

                # 2. Get main write path
                wr_path = wr_main_trace[0][1:-1]
                wr_qh = wr_path[pt-2]  # Write delay + zero based numbering compensation

                # 3. Complete guards for each write trace
                completeGuards(wr_main_trace, wr_clr_trace)

                # 4. Merge both TSs into one
                wr_trace = mergeTS(wr_main_trace, wr_clr_trace, True)

                # 5. Unroll execution conditions along the path
                wr_trace = unrollAlongPathTS(aFile, aPhaseDict, wr_trace, wr_path, "wr")

                # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
                # B. Branching trace
                # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

                # 6. Build main trace for branching instruction
                br_trace = buildCondMappedTS(aFile, aPhaseDict, br_conds,
                    "br", 1, pipe_len + hbr - 1)

                # 7. Get main branch path
                br_path = br_trace[0][1:-1]
                br_qh = br_path[pw-2]  # Write delay + zero based numbering compensation

                # 8. Unroll execution conditions along the path
                br_trace = unrollAlongPathTS(aFile, aPhaseDict, br_trace, br_path, "br")

                # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
                # C. Common for both traces
                # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

                # 7. Complete quards between branch and write trace
                info("Finalizing guards ...")
                # Add guards from other systems
                completeGuards(br_trace, wr_trace)
                #completeGuards(fi_trace, wr_trace)
                #completeGuards(fi_trace, br_trace)

                # 8. Pruning ...
                info("Pruning transition relations of transition systems ...")
                # Prune behaviour of TS according to possible stalls, conflicts
                #wr_trace = stallPrune(aFile, aPhaseDict, aReps, aUarchs, defs,
                #    aUAsserts, wr_trace, br_trace)
                # Prune behaviour of TSs according to possible clears
                (wr_trace, br_trace) = clearPrune(aFile, aPhaseDict, aReps, aUarchs, defs,
                    aUAsserts, wr_trace, br_trace, wr_path, br_path)

                # 9. Flag hazard states for writing trace
                (wr_trace, br_trace) = flagHazardousTS(wr_trace, wr_qh, hwr, "h", br_trace)

                # 10. Eliminate unreachable states
                (wr_trace, br_trace) = eliminateUnreachableTS(wr_trace, br_trace)

                # 11. Flag hazard states for branching trace
                (br_trace, wr_trace) = flagHazardousTS(br_trace, br_qh, hbr, "r", wr_trace)

                # 12. Eliminate unreachable states
                (br_trace, wr_trace) = eliminateUnreachableTS(br_trace, wr_trace)

                # Show created traces
                info("Jump trace:")
                printTS(br_trace)
                info("Write trace:")
                printTS(wr_trace)

                info("Building transducer for ARMC ...")
                M = buildTransducer([br_trace, wr_trace, fi_trace])
                #printTransducer(M)
                dotTransducer(M)

                info("Building a set of Bad states ...")
                Bad = []
                for qbr in br_trace[0]:
                    for qwr in wr_trace[0]:
                        if not re.match('r[1-9][0-9]*', qbr[2]) is None and \
                                not re.match('h[1-9][0-9]*', qwr[2]) is None:
                            Bad += [ [q2str(qbr), q2str(qwr)] ]

                print Bad

                info("Exporting problem to ARMC ...")
                con = headerARMC(aFile)
                con += "let rtl _ =\n"
                con += exportSigmaARMC(M)
                con += "in\n"
                con += exportInitARMC([br_trace, wr_trace, fi_trace])
                con += "in\n"
                con += exportTauARMC(M)
                con += "in\n"
                con += exportBadARMC(M, Bad)
                con += "in\n"
                con += exportSigmaStarARMC(M)
                con += "in\n"
                con += footerARMC()

                saveARMC(ARMC_EXPORT_FILE, con)
                info("ARMC File created: %s" % ARMC_EXPORT_FILE)

                info("Running ARMC ...")
                armc_run = runarmc(ARMC_EXPORT_FILE)
                if armc_run == ARMCHOLD:
                    info("-" * 76)
                    info("--- ARMC Result: %s" % armcrun2str(armc_run))
                    info("-" * 76)
                else:
                    error("!" * 76)
                    error("!!! ARMC Result: %s" % armcrun2str(armc_run))
                    error("!!! Combination %s @%d <--> %s (%s) @%d" % (sw, pw, st[0], st[1], pt))
                    error("!" * 76)
                    if stop_error:
                        raw_input("Error found! Press Enter to continue...")

                if debug_raw:
                    if debug_stop:
                        raw_input("Press Enter to continue...")


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Utilities
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

def getDests(aFile, aPhaseDict, aReps):
    '''
    Gets possible destinations. Consists of (storage,type) pair where type is 
    a type of incoming connection.
    '''
    dsts = []
    for phase in aPhaseDict.values():
        for item in phase:
            if isArch(aFile, getName(item), getType(item)):
                if getType(item) == TREG or (getType(item) == TMEM and getPortDir(aFile, getName(item)) == TMEMWO):
                    dsts += [ (item, CEN) ]
                    dsts += [ (item, TD) ]

                if getType(item) == TMEM and getPortDir(aFile, getName(item))== TMEMWO:
                    dsts += [ (item, TADDR) ]

    for rep in aReps.values():
        if rep != None:
            dsts += [ (rep, CCL) ]

    return dsts

def getRAWs(aFile):
    '''
    Gets all items with possible RAW hazard according to annotation.
    '''
    q = simpleQuery(aFile, "(set i where (in i (union registers memories)) (getattr i \"rawh\"))")
    return parseSet(q)

def getWAWs(aFile):
    '''
    Gets all items with possible WAW hazard according to annotation.
    '''
    q = simpleQuery(aFile, "(set i where (in i (union registers memories)) (getattr i \"wawh\"))")
    return parseSet(q)

def getWARs(aFile):
    '''
    Gets all items with possible WAW hazard according to annotation.
    '''
    q = simpleQuery(aFile, "(set i where (in i (union registers memories)) (getattr i \"warh\"))")
    return parseSet(q)

def getCTLs(aFile):
    '''
    Gets all items with possible control hazard according to annotation.
    '''
    q1 = simpleQuery(aFile, "(set i where (in i registers) (getattr i \"pc\") (getattr i \"warh\"))")
    q2 = simpleQuery(aFile, "(set i where (in i registers) (getattr i \"pc\") (getattr i \"rawh\"))")
    return list(set(parseSet(q1) + parseSet(q2)))

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Entry point
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def hazard(aFile, aUAsserts, aStat):
    info("Pre-computing often accessed data ...")

    # Determine phase and access phases
    info("    * access times of storages")
    (writes, reads, phases) = cacheAccessPhases(aFile)

    # Storages with possible RAW hazard
    info("    * storages with possible hazard")
    raws = getRAWs(aFile)
    waws = getWAWs(aFile)
    wars = getWARs(aFile)
    ctls = getCTLs(aFile)

    # Storages of each phase
    info("    * dictionary of phases")
    phase_dict = getPhaseDict(phases)

    # Compute representatives for each phase
    info("    * representatives")
    reps = getReps(aFile, phase_dict)

    # Compute possible destinations
    info("    * possible destinations")
    dsts = getDests(aFile, phase_dict, reps)

    # Determine micro-architectural registers
    info("    * micro-architectural registers")
    uarchs = getMicroArch(aFile)

    # Extract program counter
    pc = getProgramCounter(aFile)
    # FIXME: Delete ...
    #if (pc, TD) in dsts:
    #    dsts.remove((pc, TD))
    if (pc, CEN) in dsts:
        dsts.remove((pc, CEN))
    # FIXME: Delete!!!
    if (pc, TD) in dsts:
        dsts.remove((pc, TD))
    if pc in raws:
        raws.remove(pc)
    if pc in waws:
        waws.remove(pc)
    if pc in wars:
        wars.remove(pc)

    # Do inspection for each possible RAW hazard
    for raw in raws:
        inspectRAW(aFile, phases, writes, reads, phase_dict, reps, uarchs, \
            aUAsserts, dsts, raw, aStat)

    # Do inspection for each possible WAW hazard
    for waw in waws:
        inspectWAW(aFile, phases, writes, reads, phase_dict, reps, uarchs, \
            aUAsserts, waw, aStat)

    # Do inspection for each possible WAR hazard
    for war in wars:
        inspectWAR(aFile, phases, writes, reads, phase_dict, reps, uarchs, \
            aUAsserts, dsts, war, aStat)

    # Do inspection for each possible control hazard
    for ctl in ctls:
        inspectCTL(aFile, phases, writes, reads, phase_dict, reps, uarchs,
                   aUAsserts, dsts, ctl, aStat)

def hazardWrapper(aFile, aUAssertFile, aStat, aTimeStat):
    try:
        hFile = open(aFile)
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aFile)

    try:
        hFile = open(aUAssertFile)
        asserts = hFile.read()
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aUAssertFile)
        sys.exit(1)

    try:
        hFile = open(aStat, "w")
        hFile.write("HC Type;Destination;Count\n")
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aStat)
        sys.exit(1)

    try:
        hFile = open(aTimeStat, "w")
        hFile.write("Tool;Time\n")
        hFile.close()
    except IOError:
        error("Cannot open file %s!" % aTimeStat)

    hazard(aFile, asserts, aStat)

    time_stats = getTimeStatistics()
    info("Time: " + str(time_stats))
    try:
        hFile = open(aTimeStat, "a")
        hFile.write("rtlmanip;%.2f\n" % time_stats[0])
        hFile.write("smt;%.2f\n"      % time_stats[1])
        hFile.write("armc;%.2f\n"     % time_stats[2])
        hFile.close()
    except IOError:
        error("Cannot write time statitistics to %s!" % aTimeStat)

if __name__ == "__main__":
    if len(sys.argv) != 5:
        print "Usage: hazard.py anotated_vam micros_assert statistics_file time_stats_file"
        print "     anotated_vam    Filename with consistent and anotated VAM model"
        print "                     using DFA"
        print "     micros_assert   Additional asserts about microarch."
        print "                     registers."
        print "     statistics_file File with statistic measurements."
        print "     time_stats_file File with time statistics."
        sys.exit(1)

    hazardWrapper(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])
    finish()

