#!/usr/bin/python
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Name:         micro.py
# Description:  Analyzes possible states of microarchitectural registers.
# Author:       Lukas Charvat 20.3.2014
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import sys
import os

from common import *
from consistency import getMicroArch
from collections import deque

debug_micro = False
debug_stop = False
debug_dots = True

gId = 0

def printStates(aFile, aOut, aMicros, aStates):
    smt =  ";;; Enumeration of states of microarchitectural registers\n"
    smt += ";;;     {0:8} {1:8} {2:8} {3:50}\n".format("Name", "Size", "Init", "Comment")
    smt += ";;;     " + "-" * 70 + "\n"
    for micro in aMicros:
        q = simpleQuery(aFile, "(bitwidth (select register %s))" % getName(micro))
        sz = parseInt(q)
        qsig = getName(getDataout(aFile, micro))
        dsig = getName(getDatain(aFile, micro))
        init = getInitval(aFile, micro)
        comm = getComment(aFile, getName(micro), getType(micro))
        if len(comm) > 50:
            comm = "... " + comm[-46:]
        smt+= ";;;     {0:8} {1:8} {2:8} {3:50}\n".format(getName(micro), \
            str(sz), init, comm)
 
    smt += "(assert \n"

    for i in range(len(aStates)):
        state = [ (sig, aStates[i][sig]) for sig in aStates[i] ]
        ands = ""
        for j in range(len(state)):
            if j == len(state) - 1:
                ands += "             (= %s %s)\n" % state[j]
                ands += ")" * (len(state) - 1)  + "\n"
            else:
                ands += "        (and (= %s %s)\n" % state[j]

        if i == len(aStates) - 1:
            smt += "%s" % ands
            smt += ")" * (len(aStates) - 1) + "\n"
        else:
            smt += "    (or \n%s\n" % ands
    
    smt += ")\n"
    info("Number of generated states: %d" % len(aStates))

    if debug_micro:
        info("Final set of states:\n%s" % smt)

    try:
        hFile = open(aOut, "w")
        hFile.write(smtlibExport(smt))
        hFile.close()
    except IOError:
        error("Cannot open output file %s!" % aOut)

    info("Saved to: %s" % aOut)

def getInitval(aFile, aReg):
    q = simpleQuery(aFile, "(getattr (select register %s) \"init\")" % getName(aReg))
    initval = parseInt(q)

    q = simpleQuery(aFile, "(bitwidth (select register %s))" % getName(aReg))
    bw = parseInt(q)

    return "(_ bv%d %d)" % (initval, bw)

def makeId():
    global gId
    gId += 1

    return "q" + str(gId)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Utilities
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def SCC(vertices, edges):
    '''
    Find strongly connected components in graph G = (vertices, edges).

    Example:

    vertices = [1, 2, 3, 4, 5, 6, 7, 8]
    edges = {1: [2], 2: [3, 8], 3: [4, 7], 4: [5],
             5: [3, 6], 6: [], 7: [4, 6], 8: [1, 7]}
    for scc in SCC(vertices, edges):
        print(scc)

    ...  
    set([6])
    set([3, 4, 5, 7])
    set([8, 1, 2])
    '''

    identified = set()
    stack = []
    index = {}
    boundaries = []

    for v in vertices:
        if v not in index:
            to_do = [('VISIT', v)]
            while to_do:
                operation_type, v = to_do.pop()
                if operation_type == 'VISIT':
                    index[v] = len(stack)
                    stack.append(v)
                    boundaries.append(index[v])
                    to_do.append(('POSTVISIT', v))
                    # We reverse to keep the search order identical to that of
                    # the recursive code;  the reversal is not necessary for
                    # correctness, and can be omitted.
                    to_do.extend(
                        reversed([('VISITEDGE', w) for w in edges[v]]))
                elif operation_type == 'VISITEDGE':
                    if v not in index:
                        to_do.append(('VISIT', v))
                    elif v not in identified:
                        while index[v] < boundaries[-1]:
                            boundaries.pop()
                else:
                    # operation_type == 'POSTVISIT'
                    if boundaries[-1] == index[v]:
                        boundaries.pop()
                        scc = set(stack[index[v]:])
                        del stack[index[v]:]
                        identified.update(scc)
                        yield scc

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Entry point
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def micro(aFile, aOut):
    # Determine microarchitectural registers
    micros = sorted(getMicroArch(aFile))

    # Filter small micro. arch. registers only
    small_micros = []
    for micro in micros:
        q = simpleQuery(aFile, "(bitwidth (select register %s))" % getName(micro))
        sz = parseInt(q)
        if sz > 4:
            warning("Ignoring large register %s of size %d." % (getName(micro), sz))
        else:
            small_micros += [ micro ]
    
    micros = small_micros

    # Print info about discovery
    info("Detected microarchitectural registers:")
    info("    {0:8} {1:8} {2:16} {3:16} {4:8}".format("Name", "Size", "Data-in", "Data-out", "Initval"))
    info("    " + "-" * 61)
    for micro in micros:
        q = simpleQuery(aFile, "(bitwidth (select register %s))" % getName(micro))
        sz = parseInt(q)
        qsig = getName(getDataout(aFile, micro))
        dsig = getName(getDatain(aFile, micro))
        init = getInitval(aFile, micro)
        info("    {0:8} {1:8} {2:16} {3:16} {4:8}".format(getName(micro), \
            str(sz), dsig, qsig, init))
    if debug_stop:
        raw_input("Press Enter to continue...")

    # Maps D signals to Q signals of registers
    dsigs = {}
    for micro in micros:
        qsig = getName(getDataout(aFile, micro))
        dsig = getName(getDatain(aFile, micro))
        dsigs[dsig] = qsig

    # Compute init state
    state = {}
    for micro in micros:
        q = simpleQuery(aFile, "(bitwidth (select register %s))" % getName(micro))
        sz = parseInt(q)
        sig = getName(getDataout(aFile, micro))
        state[sig] = getInitval(aFile, micro)

    # Get init state short ID
    state_id = makeId()

    # Parallel composition of states of all microarch register represents
    # one state in this analysis
    states = {}
    states[state_id] = state

    # Transitions
    trans = set([])

    # State work-list
    work_list = deque([])
    workListJoin(work_list, state_id)

    while (len(work_list) > 0):
        if debug_micro:
            info("States: %s" % str(states))
        if debug_micro:
            info("Transitions: %s" % str(trans))
        if debug_micro:
            info("Worklist: %s" % str(work_list))

        state_id = work_list.popleft()
        state = states[state_id]

        coib = "(expr-coib 0 (data-in (select register %s))) "
        smt = ""
        union = ""
        for micro in micros:
            union += coib % getName(micro)

        sig_def = "(select signal %s) "
        sig_defs = ""
        for micro in micros:
            sig_defs += sig_def % getName(getDataout(aFile, micro))

        q = simpleQuery(aFile, "(smtlib (union (references (union %s)) %s))" % (union, sig_defs))
        smt += parseSmtlibExport(q) + "\n"

        q = simpleQuery(aFile, "(smtlib (union %s))" % union)
        smt += parseSmtlibExport(q) + "\n"

        for sig in state:
            smt += "(assert (= %s %s))\n" % (sig, state[sig])

        found_states = []
        while True:
            if len(found_states) > 0:
                not_assert = "(assert (not\n"
                for i in range(len(found_states)):
                    na_state = []
                    for sig in dsigs:
                        na_state += [ (sig, found_states[i][dsigs[sig]]) ]
                    ands = ""
                    for j in range(len(na_state)):
                        if j == len(na_state) - 1:
                            ands += "             (= %s %s)\n" % na_state[j]
                            ands += ")" * (len(na_state) - 1)  + "\n"
                        else:
                            ands += "        (and (= %s %s)\n" % na_state[j]
        
                    if i == len(found_states) - 1:
                        not_assert += "%s" % ands
                        not_assert += ")" * (len(found_states) - 1) + "\n"
                    else:
                        not_assert += "    (or \n%s\n" % ands
         
                not_assert += "))\n"
            else:
                not_assert = ""

            if debug_micro:
                info("Stringified version of asserts:\n%s" % str(not_assert))

            smt_check = smt + not_assert + "\n(check-sat)\n(get-model)\n(exit)\n"
                                 
            # Get model if formula is SAT
            model = smtGetBvModel(smt_check)
            if model == None:
                break

            model = dict(model)
            new_state = {}
            for sig in dsigs:
                new_state[dsigs[sig]] = model[sig]

            found_states.append(new_state)

            if not new_state in states.values():
                new_state_id = makeId()
                states[new_state_id] = new_state
                workListJoin(work_list, new_state_id)

            for (key, val) in states.items():
                if val == new_state:
                    tran = (state_id, key)
                    trans.add(tran)
                    break

            if debug_micro:
                info("States: %s" % str(states))
            if debug_micro:
                info("Transitions: %s" % str(trans))
            if debug_micro:
                info("Worklist: %s" % str(work_list))
            if debug_micro:
                info("Found states: %s" % str(found_states))
            if debug_stop:
                raw_input("Press Enter to continue...")
 
        if debug_dots:
            sys.stdout.write(".")
            sys.stdout.flush()

    if debug_dots:
        print

    vertices = states.keys()
    edges = {}
    for state in states:
        edges[state] = [ t[1] for t in trans if t[0] == state ]

    if debug_micro:
        info("Vertices: %s" % str(vertices))
        info("Edges: %s" % str(edges))
        for scc in SCC(vertices, edges):
            info("SCC: %s" % str(scc))

    print_state_ids = set([])
    for scc in SCC(vertices, edges):
        if len(scc) > 1:
            print_state_ids = print_state_ids.union(scc)
        else:
            q = list(scc)[0]
            if q in edges[q]:
                print_state_ids = print_state_ids.union(scc)

    info("States to print: %s" % str(print_state_ids))
    if debug_stop:
        raw_input("Press Enter to continue...")

    print_states = []
    for sid in print_state_ids:
        print_states.append(states[sid])

    printStates(aFile, aOut, micros, print_states)
        
def microWrapper(aFile, aOut):
    try:
        hFile = open(aFile)
        hFile.close()
        micro(aFile, aOut)
    except IOError:
        error("Cannot open file %s!" % aFile)

if __name__ == "__main__":
    if len(sys.argv) != 3:
        print "Usage: micro.py anotated_vam output_file"
        print "     anotated_vam    Filename with consistent and anotated VAM model"
        print "                     using DFA"
        print "     output_file     File to store enumuration of possible"
        print "                     states of microarch. registeres written in"
        print "                     SMT-Lib format"
        sys.exit(1)

    microWrapper(sys.argv[1], sys.argv[2])
    finish()

